from datetime import datetime
from typing import Literal
from stidantic.types import Identifier, StixLanguage


# 7.1 Language Content
class LanguageContent(StixLanguage):
    type: Literal["language-content"] = "language-content"  # pyright: ignore[reportIncompatibleVariableOverride]
    # The object_ref property identifies the id of the object that this Language Content applies to.
    # It MUST be the identifier for a STIX Object.
    object_ref: Identifier
    # The object_modified property identifies the modified time of the object that this Language Content applies to.
    # It MUST be an exact match for the modified time of the STIX Object being referenced.
    object_modified: datetime | None = None
    # The contents property contains the actual Language Content (translation).
    # The keys in the dictionary MUST be RFC 5646 language codes for which language content is being provided [RFC5646].
    # The values each consist of a dictionary that mirrors the properties in the target object
    # (identified by object_ref and object_modified). For example, to provide a translation of the name property
    # on the target object the key in the dictionary would be name.
    # For each key in the nested dictionary:
    # ●      If the original property is a string, the corresponding property in the language content object
    # MUST contain a string with the content for that property in the language of the top-level key.
    # ●      If the original property is a list, the corresponding property in the translation object must also be
    # a list. Each item in this list recursively maps to the item at the same position in the list contained in the
    # target object. The lists MUST have the same length.
    # ●      In the event that translations are only provided for some list items, the untranslated list items MUST
    # be represented by an empty string (""). This indicates to a consumer of the Language Content object that they
    # should interpolate the translated list items in the Language Content object with the corresponding (untranslated)
    # list items from the original object as indicated by the object_ref property.
    # ●      If the original property is an object (including dictionaries), the corresponding location in the
    # translation object must also be an object. Each key/value field in this object recursively maps to the object
    # with the same key in the original.
    # The translation object MAY contain only a subset of the translatable fields of the original. Keys that point to
    # non-translatable properties in the target or to properties that do not exist in the target object MUST be ignored.
    contents: dict[str, dict[str, str]]



from pydantic import ValidationInfo


def validate_identifier(value: str) -> str:
    _type, _uuid = value.split("--")
    if not _type and not _uuid:
        raise ValueError("Invalid identifier format.")
    return value


def validate_bin_field(value: str, info: ValidationInfo) -> str:
    if info.field_name and not info.field_name.endswith("_bin"):
        raise ValueError("The property name MUST end with '_bin'.")
    return value


def validate_hex_field(value: str, info: ValidationInfo) -> str:
    if info.field_name and not info.field_name.endswith("_hex"):
        raise ValueError("The property name MUST end with '_hex'.")
    return value



from enum import Enum
from re import compile
from datetime import datetime
from typing import Annotated, Literal, Any, Self, Type, get_args  # pyright: ignore[reportDeprecated]
from annotated_types import Ge, Le
from typing_extensions import TypedDict

from pydantic.functional_validators import AfterValidator, model_validator
from pydantic.networks import AnyUrl, UrlConstraints
from pydantic.types import Base64Bytes, StringConstraints
from pydantic_core import CoreSchema, core_schema
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    GetCoreSchemaHandler,
    SerializeAsAny,
)

from stidantic.validators import (
    validate_bin_field,
    validate_hex_field,
)


# Common constraints on Stix dictionnary keys.
StixKeyPattern = compile(r"^[a-zA-Z0-9\-\_]+$")
StixKeyConstraint = StringConstraints(max_length=250, pattern=StixKeyPattern)
StixKey = Annotated[str, StixKeyConstraint]

# Common constraints on Stix type names.
StixTypePattern = compile(r"^[a-zA-Z0-9\-]+$")
StixTypeConstraint = StringConstraints(pattern=StixTypePattern)
StixType = Annotated[str, StixTypeConstraint]

# Common constraints on Stix property names.
StixPropPattern = compile(r"^[a-zA-Z0-9\_]+$")
StixPropConstraint = StringConstraints(
    min_length=3, max_length=250, pattern=StixPropPattern
)
StixProp = Annotated[str, StixPropConstraint]

# 2.1 Binary
# The binary data type represents a sequence of bytes. In order to allow pattern matching on custom objects,
# for all properties that use the binary type, the property name MUST end with '_bin'.
type StixBinary = Annotated[Base64Bytes, AfterValidator(validate_bin_field)]

# 2.8 Hexadecimal
# The hex data type encodes an array of octets (8-bit bytes) as hexadecimal. The string MUST consist of an even number
# of hexadecimal characters, which are the digits '0' through '9' and the lower-case letters 'a' through 'f'. In order
# to allow pattern matching on custom objects, for all properties that use the hex type,
# the property name MUST end with '_hex'.
type StixHex = Annotated[str, AfterValidator(validate_hex_field)]

# 2.3 Dictionnary
# Dictionary keys MUST be unique in each dictionary, MUST be in ASCII, and are limited to the characters
# a-z (lowercase ASCII), A-Z (uppercase ASCII), numerals 0-9, hyphen (-), and underscore (_).
# Dictionary keys MUST be no longer than 250 ASCII characters in length and SHOULD be lowercase.
type StixDict = dict[
    StixKey,
    Any,  # pyright: ignore[reportExplicitAny]
]


# A URL reference to an external resource [RFC3986].
type StixUrl = Annotated[AnyUrl, UrlConstraints(preserve_empty_path=True)]


# 2.9 Identifier
class Identifier(str):
    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        source_type: Any,  # pyright: ignore[reportExplicitAny, reportAny]
        handler: GetCoreSchemaHandler,
    ) -> CoreSchema:
        return core_schema.no_info_after_validator_function(cls, handler(str))

    def get_type(self: str) -> str:
        return self.split("--", maxsplit=1)[0]


class StixCore(BaseModel):
    model_config = ConfigDict(  # pyright: ignore[reportUnannotatedClassAttribute]
        use_enum_values=True,
        validate_by_alias=True,
        validate_by_name=True,
        frozen=True,
        extra="allow",
    )


# 2.7 Hashes
class Hashes(StixCore, extra="allow"):
    """
    The Hashes type represents one or more cryptographic hashes, as a special set of key/value pairs.
    Accordingly, the name of each hashing algorithm MUST be specified as a key in the dictionary
    and MUST identify the name of the hashing algorithm used to generate the corresponding value.
    This name SHOULD come from one of the values defined in the hash-algorithm-ov open vocabulary.

    To enhance compatibility, the SHA-256 hash SHOULD be used whenever possible.
    """

    # Specifies the MD5 message digest algorithm. The corresponding hash string for this
    # value MUST be a valid MD5 message digest as defined in [RFC1321].
    md5: Annotated[str | None, Field(serialization_alias="MD5")] = None
    # Specifies the SHA-1 (secure-hash algorithm 1) cryptographic hash function.
    # The corresponding hash string for this value MUST be a valid SHA-1 message digest as defined in [RFC3174].
    sha1: Annotated[str | None, Field(serialization_alias="SHA-1")] = None
    # Specifies the SHA-256 cryptographic hash function (part of the SHA2 family).
    # The corresponding hash string for this value MUST be a valid SHA-256 message digest as defined in [RFC6234].
    sha256: Annotated[str | None, Field(serialization_alias="SHA-256")] = None
    # Specifies the SHA-512 cryptographic hash function (part of the SHA2 family).
    # The corresponding hash string for this value MUST be a valid SHA-512 message digest as defined in [RFC6234].
    sha512: Annotated[str | None, Field(serialization_alias="SHA-512")] = None
    # Specifies the SHA3-256 cryptographic hash function. The corresponding hash string
    # for this value MUST be a valid SHA3-256 message digest as defined in [FIPS202].
    sha3_256: Annotated[str | None, Field(serialization_alias="SHA3-256")] = None
    # Specifies the SHA3-512 cryptographic hash function. The corresponding hash string
    # for this value MUST be a valid SHA3-512 message digest as defined in [FIPS202].
    sha3_512: Annotated[str | None, Field(serialization_alias="SHA3-512")] = None
    # Specifies the ssdeep fuzzy hashing algorithm. The corresponding hash string for this
    # value MUST be a valid piecewise hash as defined in the [SSDEEP] specification.
    ssdeep: Annotated[str | None, Field(serialization_alias="SSDEEP")] = None
    # Specifies the TLSH fuzzy hashing algorithm. The corresponding hash string for this
    # value MUST be a valid 35 byte long hash as defined in the [TLSH] specification.
    tlsh: Annotated[str | None, Field(serialization_alias="TLSH")] = None

    @model_validator(mode="after")
    def lang_or_marking_ref(self) -> "Hashes":
        """
        Dictionary keys MUST be unique in each hashes property, MUST be in ASCII, and are limited to the
        characters a-z (lowercase ASCII), A-Z (uppercase ASCII), numerals 0-9, hyphen (-), and underscore (_).
        Dictionary keys MUST have a minimum length of 3 ASCII characters
        and MUST be no longer than 250 ASCII characters in length.
        The value MUST be a string in the appropriate format defined by the hash type indicated in the dictionary key.
        """
        if self.__pydantic_extra__ and any(
            not (len(key) > 250 or len(key) < 3 or StixKeyPattern.match(key))
            for key in self.__pydantic_extra__.keys()
        ):
            raise ValueError("Invalid extra hash key.")
        return self


# 2.5 External Reference
class ExternalReference(StixCore):
    """
    External references are used to describe pointers to information represented outside of STIX.
    For example, a Malware object could use an external reference to indicate an ID for that malware
    in an external database or a report could use references to represent source material.
    """

    # The name of the source that the external-reference is defined within (system, registry, organization, etc.).
    source_name: str
    # A human readable description.
    description: str | None = None
    # A URL reference to an external resource [RFC3986].
    url: StixUrl | None = None
    # Specifies a dictionary of hashes for the contents of the url. This SHOULD be provided when the url property is
    # present. Dictionary keys MUST come from one of the entries listed in the hash-algorithm-ov open vocabulary.
    # As stated in Section 2.7, to ensure interoperability, a SHA-256 hash SHOULD be included whenever possible.
    hashes: Hashes | None = None
    # An identifier for the external reference content.
    external_id: str | None = None

    @model_validator(mode="before")
    def at_least_one(cls, values: dict[str, str]) -> dict[str, str]:
        """
        In addition to the source_name property, at least one of the description, url,
        or external_id properties MUST be present.
        """
        for value in values.values():
            if value:
                return values
        raise ValueError("Missing at least one hash value.")


# 2.11 Kill Chain Phase
class KillChainPhase(StixCore):
    """
    The kill-chain-phase represents a phase in a kill chain, which describes the
    various phases an attacker may undertake in order to achieve their objectives.
    """

    # The name of the kill chain. The value of this property SHOULD be all lowercase
    # and SHOULD use hyphens instead of spaces or underscores as word separators.
    kill_chain_name: str
    # The name of the phase in the kill chain. The value of this property SHOULD be all
    # lowercase and SHOULD use hyphens instead of spaces or underscores as word separators.
    phase_name: str


# 7.2.3 Granular Markings
class GranularMarking(StixCore):
    """
    The granular-marking type defines how the list of marking-definition objects referenced by
    the marking_refs property to apply to a set of content identified by
    the list of selectors in the selectors property.
    """

    # The lang property identifies the language of the text identified by this marking. The value of the lang property,
    # if present, MUST be an [RFC5646] language code. If the marking_ref property is not present, this property MUST
    # be present. If the marking_ref property is present, this property MUST NOT be present.
    lang: str | None = None
    # The marking_ref property specifies the ID of the marking-definition object that describes the marking.
    # If the lang property is not present, this property MUST be present. If the lang property is present,
    # this property MUST NOT be present.
    marking_ref: Identifier | None = None
    # The selectors property specifies a list of selectors for content contained within the STIX Object in which this
    # property appears. Selectors MUST conform to the syntax defined below. The marking-definition referenced in
    # the marking_ref property is applied to the content selected by the selectors in this list. The [RFC5646] language
    # code specified by the lang property is applied to the content selected by the selectors in this list.
    selectors: list[str] | None = None
    # Selector Syntax:
    # Selectors contained in the selectors list are strings that consist of multiple components that MUST be separated
    # by the . character. Each component MUST be one of:
    #    ●      A property name or dictionary key, e.g., description, or;
    #    ●      A zero-based list index, specified as a non-negative integer in square brackets, e.g., [4]
    # Selectors denote path traversals: the root of each selector is the STIX Object
    # that the granular_markings property appears in.
    # Starting from that root, for each component in the selector, properties and list items are traversed.
    # When the complete list has been traversed, the value of the content is considered selected.
    # Selectors MUST refer to properties or list items that are actually present on the marked object.

    @model_validator(mode="after")
    def lang_or_marking_ref(self) -> "GranularMarking":
        if self.lang and self.marking_ref:
            raise ValueError(
                "Both lang and marking_ref properties MUST NOT be present."
            )
        if not self.lang and not self.marking_ref:
            raise ValueError("Either lang or marking_ref property MUST be present.")
        return self


# 10.5 Extension Types Enumeration
class ExtensionType(Enum):
    new_sdo = "new-sdo"
    new_sco = "new-sco"
    new_sro = "new-sro"
    property_extension = "property-extension"
    toplevel_property_extension = "toplevel-property-extension"


class Extension(StixCore):
    extension_type: ExtensionType | None = None


class ExtensionsDict(TypedDict, total=False, extra_items=SerializeAsAny[Extension]): ...  # noqa: E701


# 3.2 Common Properties
class StixCommon(StixCore):
    """
    This section defines the common properties that MAY exist on a STIX Objects. While some STIX Objects use all
    of these common properties, not all object types do.
    Each type of STIX Object defines which common properties are required, which are optional,
    and which are not in use.
    """

    # All STIX Objects and the STIX Bundle Object have an id property that uniquely identifies each instance
    # of the object.
    # This id MUST meet the requirements of the identifier type (see section 2.9).
    # For objects that support versioning, all objects with the same id are considered different versions
    # of the same object and the version of the object is identified by its modified property.
    id: Identifier
    # The type property identifies the type of STIX Object.
    type: str
    # The value of this property MUST be 2.1 for STIX Objects defined according to this specification.
    spec_version: Literal["2.1"] = "2.1"
    # The revoked property is only used by STIX Objects that support versioning and indicates whether the object
    # has been revoked.
    revoked: bool | None = None
    # The labels property specifies a set of terms used to describe this object.
    # Where an object has a specific property defined in the specification for characterizing
    # subtypes of that object, the labels property MUST NOT be used for that purpose.
    labels: list[str] | None = None
    # The confidence property identifies the confidence that the creator has in the correctness of their data.
    # The confidence value MUST be a number in the range of 0-100.
    confidence: Annotated[int, Ge(0), Le(100)] | None = None
    # The lang property identifies the language of the text content in this object.
    # When present, it MUST be a language code conformant to [RFC5646].
    # RFC5646 does not enforce ISO 639-1 alpha-2 or ISO 639-3 alpha-3 formats.
    lang: str | None = None
    # The external_references property specifies a list of external references which refers to non-STIX information.
    external_references: list[ExternalReference] | None = None
    # The object_marking_refs property specifies a list of id properties of marking-definition objects that apply
    # to this object.
    object_marking_refs: list[Identifier] | None = None
    # The granular_markings property specifies a list of granular markings applied to this object.
    granular_markings: list[GranularMarking] | None = None
    # Specifies any extensions of the object, as a dictionary.
    # Dictionary keys SHOULD be the id of a STIX Extension object or the name of a predefined object extension found
    # in this specification, depending on the type of extension being used.
    # The corresponding dictionary values MUST contain the contents of the extension instance.
    # Each extension dictionary MAY contain the property extension_type.
    # The value of this property MUST come from the extension-type-enum enumeration.
    # If the extension_type property is not present, then this is a predefined extension which
    # does not use the extension facility described in section 7.3.
    # When this extension facility is used the extension_type property MUST be present.
    extensions: ExtensionsDict | dict[str, Extension] | None = None

    @classmethod
    def register_new_extension(
        cls,
        definition: "StixCommon",  # pyright: ignore[reportUnusedParameter]
        extension: Type[Extension],  # pyright: ignore[reportUnusedParameter, reportDeprecated]
    ):
        """
        Dynamically update the extensions property so that new classes get deserialized based on the extension key.
        """
        annotations = get_args(cls.model_fields["extensions"].annotation)[0]  # pyright: ignore[reportAny, reportUnusedVariable]
        CustomExtensionsDict = TypedDict(
            "CustomExtensionsDict",
            {
                definition.id: extension,  # pyright: ignore[reportGeneralTypeIssues]
                **annotations.__annotations__,  # pyright: ignore[reportGeneralTypeIssues]
            },
            total=False,
            extra_items=SerializeAsAny[Extension],
        )
        cls.model_fields["extensions"].annotation = CustomExtensionsDict | None  # pyright: ignore[reportAttributeAccessIssue]
        cls.model_rebuild(force=True)  # pyright: ignore[reportUnusedCallResult]


class StixDomain(StixCommon):
    # The created_by_ref property specifies the id property of the identity object that describes the entity that
    # created this object.
    created_by_ref: Identifier | None = None
    # The created property represents the time at which the object was originally created.
    # The created property MUST NOT be changed when creating a new version of the object.
    created: datetime
    # The modified property is only used by STIX Objects that support versioning and represents
    # the time that this particular version of the object was last modified.
    # Object creators MUST set the modified property when creating a new
    # version of an object if the created property was set.
    modified: datetime

    @model_validator(mode="after")
    def validate_modified_after_created(self) -> Self:
        """
        If the created property is defined, then the value of the modified property for a given object version
        MUST be later than or equal to the value of the created property.
        """
        if self.created > self.modified:
            raise ValueError(
                "The modified property MUST be later than or equal to the value of the created property."
            )
        return self


class StixObservable(StixCommon):
    # This property defines whether or not the data contained within the object has been defanged.
    defanged: bool | None = None


class StixRelationship(StixCommon):
    relationship_type: str
    # The created property represents the time at which the object was originally created.
    # The created property MUST NOT be changed when creating a new version of the object.
    created: datetime
    # The modified property is only used by STIX Objects that support versioning and represents
    # the time that this particular version of the object was last modified.
    # Object creators MUST set the modified property when creating a new
    # version of an object if the created property was set.
    modified: datetime
    # The created_by_ref property specifies the id property of the identity object that describes the entity that
    # created this object.
    created_by_ref: Identifier | None = None

    @model_validator(mode="after")
    def validate_modified_after_created(self) -> Self:
        """
        If the created property is defined, then the value of the modified property for a given object version
        MUST be later than or equal to the value of the created property.
        """
        if self.created > self.modified:
            raise ValueError(
                "The modified property MUST be later than or equal to the value of the created property."
            )
        return self


class StixMeta(StixCommon):
    type: str
    # The created property represents the time at which the object was originally created.
    # The created property MUST NOT be changed when creating a new version of the object.
    created: datetime


class StixLanguage(StixMeta):
    # The created_by_ref property specifies the id property of the identity object that describes the entity that
    # created this object.
    created_by_ref: Identifier | None = None
    # The modified property is only used by STIX Objects that support versioning and represents
    # the time that this particular version of the object was last modified.
    # Object creators MUST set the modified property when creating a new
    # version of an object if the created property was set.
    modified: datetime

    @model_validator(mode="after")
    def validate_modified_after_created(self) -> Self:
        """
        If the created property is defined, then the value of the modified property for a given object version
        MUST be later than or equal to the value of the created property.
        """
        if self.created > self.modified:
            raise ValueError(
                "The modified property MUST be later than or equal to the value of the created property."
            )
        return self


class StixMarking(StixMeta):
    # The created_by_ref property specifies the id property of the identity object that describes the entity that
    # created this object.
    created_by_ref: Identifier | None = None


class StixExtension(StixMeta):
    # The created_by_ref property specifies the id property of the identity object that describes the entity that
    # created this object.
    created_by_ref: Identifier
    # The modified property is only used by STIX Objects that support versioning and represents
    # the time that this particular version of the object was last modified.
    # Object creators MUST set the modified property when creating a new
    # version of an object if the created property was set.
    modified: datetime

    @model_validator(mode="after")
    def validate_modified_after_created(self) -> Self:
        """
        If the created property is defined, then the value of the modified property for a given object version
        MUST be later than or equal to the value of the created property.
        """
        if self.created > self.modified:
            raise ValueError(
                "The modified property MUST be later than or equal to the value of the created property."
            )
        return self



from typing import Literal, Annotated, Self
from pydantic import Field
from pydantic.functional_validators import model_validator
from stidantic.types import StixExtension, ExtensionType, StixProp


# 7.3 Extension Definition
class ExtensionDefinition(StixExtension):
    """
    The STIX Extension Definition object allows producers of threat intelligence to extend existing STIX objects or
    to create entirely new STIX objects in a standardized way. This object contains detailed information about the
    extension and any additional properties and or objects that it defines. This extension mechanism MUST NOT be used
    to redefine existing standardized objects or properties.

    If a producer does not include the STIX Extension Definition object with the STIX objects that use it,
    consumers should refer to section 3.3 for information in resolving references.

    There are three ways to extend STIX using STIX Extensions.
        - Define one or more new STIX Object types.
        - Define additional properties for an existing STIX Object type as a nested property extension. This is
        typically done to represent a sub-component or module of one or more STIX Object types.
        - Define additional properties for an existing STIX Object type at the object's top-level. This can be done to
        represent properties that form an inherent part of the definition of an object type.

    When defining a new STIX Object (e.g., SDO, SCO, or SRO) all common properties associated with that type of
    object (SDO, SCO, SRO) MUST be included in the schema or definition of that new STIX Object type.
    Extensions that create new STIX objects MUST follow all conformance requirements for that object type
    (SDO, SCO, SRO) including all of the requirements for the common properties associated with that object type.

    When defining a STIX extension using the nested property extension mechanism the extensions property MUST include
    the extension definition's UUID that defines the extension definition object and the extension_type property
    as defined in section 3.2.

    IMPORTANT NOTE: Producers using top-level property extensions should be mindful that another producer could also
    define a top-level property extension using the same property names but for different purposes causing name
    conflicts when both extensions are used in the same environment. This standard does not define any name conflict
    resolution for new STIX Objects or for top-level properties created by this extension mechanism. However,
    producers SHOULD follow industry best practices such as using unique property names that are guaranteed to avoid
    duplicates across all organizations to avoid naming conflicts.
    IMPORTANT NOTE: Producers using STIX extensions should be mindful that future versions of the STIX specification
    MAY define objects and or properties that conflict with existing non-standardized extensions. In these cases the
    meaning as defined in the STIX specification will override any and all conflicting extensions.

    Specific extensions, as with specific Custom Properties, MAY NOT be supported across implementations.
    A consumer that receives STIX content containing a STIX extension that it does not understand MAY refuse to
    process the content or MAY ignore that extension and continue processing the content.

    The 3 uses of this extension facility MAY be combined into a single Extension Definition object when appropriate.

    The following example highlights where this may be useful.

    Hybrid Extension Example
    An intelligence producer has a monitoring network of sensors that collect a variety of cybersecurity  telemetry
    from each sensor where those sensors have unique data not currently defined in STIX 2.1.
    The producer wishes to create an extension that other downstream consumers can receive both the high-level
    summarization object but also the individual categorized telemetry from each sensor.
    a)    A new SDO representing the statistical summarization object.
    b)    A list of new properties to be added to the standard Observed Data object representing additional meta-data
    information associated with the telemetry.
    c)     A new SCO representing a new cyber observable data type.

    In this case, the producer creates a single extension that contains the following extension types:
    "extension_types": [ "new-sdo", "new-sco", "property-extension" ]

    Therefore, producers MAY use the hybrid extension mechanism when they wish to define a single extension that
    encompasses new SDO and/or sub-component or top-level property extension properties in a related extension.

    Producers SHOULD NOT use the hybrid extension mechanism if the extensions are not related to each other.
    If the extensions are independent features then a producer SHOULD consider creating separate extension definitions.
    """

    type: Literal["extension-definition"] = "extension-definition"  # pyright: ignore[reportIncompatibleVariableOverride]
    # A name used for display purposes during execution, development, or debugging.
    name: str
    # A detailed explanation of what data the extension conveys and how it is intended to be used.
    # While the description property is optional this property SHOULD be populated.
    # Note that the schema property is the normative definition of the extension, and this property, if present,
    # is for documentation purposes only.
    description: str | None = None
    # The normative definition of the extension, either as a URL or as plain text explaining the definition.
    # A URL SHOULD point to a JSON schema or a location that contains information about the schema
    json_schema: Annotated[str, Field(alias="schema")]
    # The version of this extension. Producers of STIX extensions are encouraged to follow standard semantic
    # versioning procedures where the version number follows the pattern, MAJOR.MINOR.PATCH. This will allow
    # consumers to distinguish between the three different levels of compatibility typically identified by
    # such versioning strings.
    version: str
    # This property specifies one or more extension types contained within this extension.
    # When this property includes toplevel-property-extension then the extension_properties property
    # SHOULD include one or more property names.
    extension_types: list[ExtensionType]
    # This property contains the list of new property names that are added to an object by an extension.
    # This property MUST only be used when the extension_types property includes a value of toplevel-property-extension.
    # In other words, when new properties are being added at the top-level of an existing object.
    # The property names used in Extension STIX Object MUST be in ASCII and MUST only contain the characters a–z
    # (lowercase ASCII), 0–9, and underscore (_).
    # The name of a property of a Extension STIX Object MUST have a minimum length of 3 ASCII characters.
    # The name of a property of a Extension STIX Object MUST be no longer than 250 ASCII characters in length.
    extension_properties: list[StixProp] | None = None

    @model_validator(mode="after")
    def validate_extension_properties(self) -> Self:
        """
        extension_properties MUST only be used when the extension_types property includes a value of
        toplevel-property-extension. In other words, when new properties are being added at the
        top-level of an existing object.
        """
        if (
            self.extension_properties
            and ExtensionType.toplevel_property_extension not in self.extension_types
        ):
            raise ValueError(
                "extension_types property can't be used without toplevel-property-extension in extension_types."
            )
        return self



from datetime import datetime
from typing import Any, Literal, Annotated, Self
from typing_extensions import deprecated
from annotated_types import Ge, Gt, Le

from pydantic import Field
from pydantic.functional_validators import model_validator

from stidantic.types import Identifier, StixDomain, KillChainPhase, ExternalReference
from stidantic.vocab import OpinionEnum


# 4.1 Attack Pattern
class AttackPattern(StixDomain):
    """
    Attack Patterns are a type of TTP that describe ways that adversaries attempt to compromise targets.

    Attack Patterns are used to help categorize attacks, generalize specific attacks to the patterns that they follow,
    and provide detailed information about how attacks are performed. An example of an attack pattern is
    "spear phishing": a common type of attack where an attacker sends a carefully crafted e-mail message
    to a party with the intent of getting them to click a link or open an attachment to deliver malware.

    Attack Patterns can also be more specific; spear phishing as practiced by a particular threat actor
    (e.g., they might generally say that the target won a contest) can also be an Attack Pattern.
    """

    type: Literal["attack-pattern"] = "attack-pattern"  # pyright: ignore[reportIncompatibleVariableOverride]
    # The name used to identify the Attack Pattern.
    name: str
    # A description that provides more details and context about the Attack Pattern,
    # potentially including its purpose and its key characteristics.
    description: str | None = None
    # Alternative names used to identify this Attack Pattern.
    aliases: list[str] | None = None
    # The list of kill chain phases for which this attack pattern is used.
    kill_chain_phases: list[KillChainPhase] | None = None


# 4.2 Campaign
class Campaign(StixDomain):
    """
    A Campaign is a grouping of adversarial behaviors that describes a set of malicious activities or attacks
    (sometimes called waves) that occur over a period of time against a specific set of targets.
    Campaigns usually have well defined objectives and may be part of an Intrusion Set.

    Campaigns are often attributed to an intrusion set and threat actors. The threat actors may reuse known
    infrastructure from the intrusion set or may set up new infrastructure specific for conducting that campaign.

    Campaigns can be characterized by their objectives and the incidents they cause, people or resources they target,
    and the resources (infrastructure, intelligence, Malware, Tools, etc.) they use.

    For example, a Campaign could be used to describe a crime syndicate's attack using a specific variant of
    malware and new C2 servers against the executives of ACME Bank during the summer of 2016 in order
    to gain secret information about an upcoming merger with another bank.
    """

    type: Literal["campaign"] = "campaign"  # pyright: ignore[reportIncompatibleVariableOverride]
    # A name used to identify the Campaign.
    name: str
    # A description that provides more details and context about the Campaign,
    # potentially including its purpose and its key characteristics.
    description: str | None = None
    # Alternative names used to identify this Campaign.
    aliases: list[str] | None = None
    # The time that this Campaign was first seen.
    # A summary property of data from sightings and other data that may or may not be available in STIX.
    # If new sightings are received that are earlier than the first seen timestamp,
    # the object may be updated to account for the new data.
    first_seen: datetime | None = None
    # The time that this Campaign was last seen.
    # A summary property of data from sightings and other data that may or may not be available in STIX.
    # If new sightings are received that are later than the last seen timestamp,
    # the object may be updated to account for the new data.
    last_seen: datetime | None = None
    # The Campaign’s primary goal, objective, desired outcome, or intended effect
    # — what the Threat Actor or Intrusion Set hopes to accomplish with this Campaign.
    objective: str | None = None

    @model_validator(mode="after")
    def validate_last_seen_after_first_seen(self) -> Self:
        """
        If the last_seen property and the first_seen property are both defined, then the last_seen property
        MUST be greater than or equal to the timestamp in the first_seen property.
        """
        if self.first_seen and self.last_seen and self.first_seen > self.last_seen:
            raise ValueError(
                "The last_seen property MUST be greater than or equal to the timestamp in the first_seen property"
            )
        return self


# 4.3 Course of Action
class CourseOfAction(StixDomain):
    """
    Note: The Course of Action object in STIX 2.1 is a stub. It is included to support basic use cases
    (such as sharing prose courses of action) but does not support the ability to represent automated courses of
    action or contain properties to represent metadata about courses of action. Future STIX 2 releases will expand it
    to include these capabilities.

    A Course of Action is an action taken either to prevent an attack or to respond to an attack that is in progress.
    It may describe technical, automatable responses (applying patches, reconfiguring firewalls) but can also describe
    higher level actions like employee training or policy changes. For example, a course of action to mitigate a
    vulnerability could describe applying the patch that fixes it.

    The Course of Action SDO contains a textual description of the action; a reserved action property also serves as a
    placeholder for future inclusion of machine automatable courses of action.
    """

    type: Literal["course-of-action"] = "course-of-action"  # pyright: ignore[reportIncompatibleVariableOverride]
    # A name used to identify the Course of Action.
    name: str
    # A description that provides more details and context about the Course of Action,
    # potentially including its purpose and its key characteristics.
    description: str | None = None
    # A reserved property that serves as a placeholder for future inclusion of machine automatable courses of action.
    # action: str | None = None


# 4.4 Grouping
class Grouping(StixDomain):
    """
    A Grouping object explicitly asserts that the referenced STIX Objects have a shared context, unlike a STIX Bundle
    (which explicitly conveys no context). A Grouping object should not be confused with an intelligence product,
    which should be conveyed via a STIX Report.

    A STIX Grouping object might represent a set of data that, in time, given sufficient analysis, would mature to
    convey an incident or threat report as a STIX Report object. For example, a Grouping could be used to characterize
    an ongoing investigation into a security event or incident. A Grouping object could also be used to assert that the
    referenced STIX Objects are related to an ongoing analysis process, such as when a threat analyst is collaborating
    with others in their trust community to examine a series of Campaigns and Indicators. The Grouping SDO contains a
    list of references to SDOs, SCOs, SROs, and SMOs, along with an explicit statement of the context shared by the
    content, a textual description, and the name of the grouping.
    """

    type: Literal["grouping"] = "grouping"  # pyright: ignore[reportIncompatibleVariableOverride]
    # A name used to identify the Grouping.
    name: str | None = None
    # A description that provides more details and context about the Grouping, potentially including its purpose and
    # its key characteristics.
    description: str | None = None
    # A short descriptor of the particular context shared by the content referenced by the Grouping.
    # The value for this property SHOULD come from the grouping-context-ov open vocabulary.
    context: str
    # Specifies the STIX Objects that are referred to by this Grouping.
    object_refs: list[Identifier]


# 4.5 Identity
class Identity(StixDomain):
    """
    Identities can represent actual individuals, organizations, or groups (e.g., ACME, Inc.) as well as classes of
    individuals, organizations, systems or groups (e.g., the finance sector).

    The Identity SDO can capture basic identifying information, contact information, and the sectors that the Identity
    belongs to. Identity is used in STIX to represent, among other things, targets of attacks, information sources,
    object creators, and threat actor identities.
    """

    type: Literal["identity"] = "identity"  # pyright: ignore[reportIncompatibleVariableOverride]
    # A name used to identify the Identity. When referring to a specific entity (e.g., an individual or organization),
    # this property SHOULD contain the canonical name of the specific entity.
    name: str
    # A description that provides more details and context about the Identity,
    # potentially including its purpose and its key characteristics.
    description: str | None = None
    # The list of roles that this Identity performs (e.g., CEO, Domain Administrators, Doctors, Hospital, or Retailer).
    # No open vocabulary is yet defined for this property.
    roles: list[str] | None = None
    # The type of entity that this Identity describes, e.g., an individual or organization.
    # The value for this property SHOULD come from the identity-class-ov open vocabulary.
    identity_class: str | None = None
    # The list of industry sectors that this Identity belongs to.
    # The value for this property SHOULD come from the industry-sector-ov open vocabulary.
    sectors: list[str] | None = None
    # The contact information (e-mail, phone number, etc.) for this Identity.
    # No format for this information is currently defined by this specification.
    contact_information: str | None = None


# 4.6 Incident
class Incident(StixDomain):
    """
    NOTE: The Incident object in STIX 2.1 is a stub. It is included to support basic use cases but does not contain
    properties to represent metadata about incidents. Future STIX 2 releases will expand it to include these
    capabilities. It is suggested that it is used as an extension point for an Incident object defined using the
    extension facility described in section 7.3.
    """

    type: Literal["incident"] = "incident"  # pyright: ignore[reportIncompatibleVariableOverride]
    # A name used to identify the Incident.
    name: str
    # A description that provides more details and context about the Incident,
    # potentially including its purpose and its key characteristics.
    description: str | None = None


class Indicator(StixDomain):
    """
    Indicators contain a pattern that can be used to detect suspicious or malicious cyber activity. For example, an
    Indicator may be used to represent a set of malicious domains and use the STIX Patterning Language
    (see section 9) to specify these domains.

    The Indicator SDO contains a simple textual description, the Kill Chain Phases that it detects behavior in, a time
    window for when the Indicator is valid or useful, and a required pattern property to capture a structured detection
    pattern. Conforming STIX implementations MUST support the STIX Patterning Language as defined in section 9.

    Relationships from the Indicator can describe the malicious or suspicious behavior that it directly detects
    (Malware, Tool, and Attack Pattern). In addition, it may also imply the presence of a Campaigns, Intrusion Sets,
    and Threat Actors, etc.
    """

    type: Literal["indicator"] = "indicator"  # pyright: ignore[reportIncompatibleVariableOverride]
    # A name used to identify the Indicator.
    # Producers SHOULD provide this property to help products and analysts understand what this Indicator actually does.
    name: str | None = None
    # A description that provides more details and context about the Indicator,
    # potentially including its purpose and its key characteristics.
    # Producers SHOULD provide this property to help products and analysts understand what this Indicator actually does.
    description: str | None = None
    # A set of categorizations for this indicator.
    # The values for this property SHOULD come from the indicator-type-ov open vocabulary.
    indicator_types: list[str] | None = None
    # The detection pattern for this Indicator MAY be expressed as a STIX Pattern as specified in section 9 or another
    # appropriate language such as SNORT, YARA, etc.
    pattern: str
    # The pattern language used in this indicator.
    # The value for this property SHOULD come from the pattern-type-ov open vocabulary.
    # The value of this property MUST match the type of pattern data included in the pattern property.
    pattern_type: str
    # The version of the pattern language that is used for the data in the pattern property which MUST match the type
    # of pattern data included in the pattern property.
    # For patterns that do not have a formal specification, the build or code version that the pattern is known to
    # work with SHOULD be used.
    # For the STIX Pattern language, the default value is determined by the specification version of the object.
    # For other languages, the default value SHOULD be the latest version of the patterning language at the time of
    # this object’s creation.
    pattern_version: str | None = None
    # The time from which this Indicator is considered a valid indicator of the behaviors it is related or represents.
    valid_from: datetime
    # The time at which this Indicator should no longer be considered a valid indicator of the behaviors it is
    # related to or represents.
    # If the valid_until property is omitted, then there is no constraint on the latest time for which the
    # Indicator is valid.
    valid_until: datetime | None = None
    # The kill chain phase(s) to which this Indicator corresponds.
    kill_chain_phases: list[KillChainPhase] | None = None

    @model_validator(mode="after")
    def validate_valid_until_after_valid_from(self) -> Self:
        """
        The valid_until property MUST be greater than the timestamp in the valid_from property.
        """
        if self.valid_from and self.valid_until and self.valid_from > self.valid_until:
            raise ValueError(
                "The valid_until property MUST be greater than the timestamp in the valid_from property"
            )
        return self


# 4.8 Infrastructure
class Infrastructure(StixDomain):
    """
    The Infrastructure SDO represents a type of TTP and describes any systems, software services and any associated
    physical or virtual resources intended to support some purpose (e.g., C2 servers used as part of an attack,
    device or server that are part of defense, database servers targeted by an attack, etc.). While elements of an
    attack can be represented by other SDOs or SCOs, the Infrastructure SDO represents a named group of related data
    that constitutes the infrastructure.
    """

    type: Literal["infrastructure"] = "infrastructure"  # pyright: ignore[reportIncompatibleVariableOverride]
    # A name or characterizing text used to identify the Infrastructure.
    name: str
    # A description that provides more details and context about the Infrastructure,
    # potentially including its purpose, how it is being used, how it relates to other intelligence activities
    # captured in related objects, and its key characteristics.
    description: str | None = None
    # The type of infrastructure being described.
    # The values for this property SHOULD come from the infrastructure-type-ov open vocabulary.
    infrastructure_types: list[str] | None = None
    # Alternative names used to identify this Infrastructure.
    aliases: list[str] | None = None
    # The list of Kill Chain Phases for which this Infrastructure is used.
    kill_chain_phases: list[KillChainPhase] | None = None
    # The time that this Infrastructure was first seen performing malicious activities.
    first_seen: datetime | None = None
    # The time that this Infrastructure was last seen performing malicious activities.
    last_seen: datetime | None = None

    @model_validator(mode="after")
    def validate_last_seen_after_first_seen(self) -> Self:
        """
        If the last_seen and the first_seen properties are both defined, then the last_seen property
        MUST be greater than or equal to the timestamp in the first_seen property.
        """
        if self.first_seen and self.last_seen and self.first_seen > self.last_seen:
            raise ValueError(
                "The last_seen property MUST be greater than or equal to the timestamp in the first_seen property"
            )
        return self


# 4.9 Intrusion Set
class IntrusionSet(StixDomain):
    """
    An Intrusion Set is a grouped set of adversarial behaviors and resources with common properties that is believed to
    be orchestrated by a single organization. An Intrusion Set may capture multiple Campaigns or other activities that
    are all tied together by shared attributes indicating a commonly known or unknown Threat Actor. New activity can be
    attributed to an Intrusion Set even if the Threat Actors behind the attack are not known. Threat Actors can move
    from supporting one Intrusion Set to supporting another, or they may support multiple Intrusion Sets.

    Where a Campaign is a set of attacks over a period of time against a specific set of targets to achieve some
    objective, an Intrusion Set is the entire attack package and may be used over a very long period of time in
    multiple Campaigns to achieve potentially multiple purposes.

    While sometimes an Intrusion Set is not active, or changes focus, it is usually difficult to know if it has
    truly disappeared or ended. Analysts may have varying level of fidelity on attributing an Intrusion Set back to
    Threat Actors.
    """

    type: Literal["intrusion-set"] = "intrusion-set"  # pyright: ignore[reportIncompatibleVariableOverride]
    # A name or characterizing text used to identify the Intrusion Set.
    name: str
    # A description that provides more details and context about the Intrusion Set,
    # potentially including its purpose and its key characteristics.
    description: str | None = None
    # Alternative names used to identify this Intrusion Set.
    aliases: list[str] | None = None
    # The time that this Intrusion Set was first seen.
    # A summary property of data from sightings and other data that may or may not be available in STIX.
    # If new sightings are received that are earlier than the first seen timestamp, the object may be updated to
    # account for the new data.
    first_seen: datetime | None = None
    # The time that this Intrusion Set was last seen.
    # This property is a summary property of data from sightings and other data that may or may not be available.
    # If new sightings are received that are later than the last seen timestamp, the object may be updated to
    # account for the new data.
    last_seen: datetime | None = None
    # The high-level goals of this Intrusion Set, namely, what are they trying to do.
    # For example, they may be motivated by personal gain, but their goal is to steal credit card numbers.
    # To do this, they may execute specific Campaigns that have detailed objectives like compromising point of sale
    # systems at a large retailer.
    # Another example: to gain information about latest merger and IPO information from ACME Bank.
    goals: list[str] | None = None
    # This property specifies the organizational level at which this Intrusion Set typically works, which in turn
    # determines the resources available to this Intrusion Set for use in an attack.
    # The value for this property SHOULD come from the attack-resource-level-ov open vocabulary.
    resource_level: str | None = None
    # The primary reason, motivation, or purpose behind this Intrusion Set.
    # The value for this property SHOULD come from the attack-motivation-ov open vocabulary.
    primary_motivation: str | None = None
    # The secondary reasons, motivations, or purposes behind this Intrusion Set.
    # The values for this property SHOULD come from the attack-motivation-ov open vocabulary.
    secondary_motivations: list[str] | None = None

    @model_validator(mode="after")
    def validate_last_seen_after_first_seen(self) -> Self:
        """
        If the last_seen and the first_seen properties are both defined, then the last_seen property
        MUST be greater than or equal to the timestamp in the first_seen property.
        """
        if self.first_seen and self.last_seen and self.first_seen > self.last_seen:
            raise ValueError(
                "The last_seen property MUST be greater than or equal to the timestamp in the first_seen property"
            )
        return self


# 4.10 Location
class Location(StixDomain):
    """
    A Location represents a geographic location. The location may be described as any, some or all of the following:
    region (e.g., North America), civic address (e.g. New York, US), latitude and longitude.

    Locations are primarily used to give context to other SDOs. For example, a Location could be used in a
    relationship to describe that the Bourgeois Swallow intrusion set originates from Eastern Europe.

    The Location SDO can be related to an Identity or Intrusion Set to indicate that the identity or intrusion set is
    located in that location. It can also be related from a malware or attack pattern to indicate that they target
    victims in that location. The Location object describes geographic areas, not governments, even in cases where that
    area might have a government. For example, a Location representing the United States describes the United States
    as a geographic area, not the federal government of the United States.
    """

    type: Literal["location"] = "location"  # pyright: ignore[reportIncompatibleVariableOverride]
    # A name used to identify the Location.
    name: str | None = None
    # A textual description of the Location.
    description: str | None = None
    # The latitude of the Location in decimal degrees.
    # Positive numbers describe latitudes north of the equator,
    # and negative numbers describe latitudes south of the equator
    latitude: Annotated[float, Ge(-90.0), Le(90.0)] | None = None
    # The longitude of the Location in decimal degrees.
    # Positive numbers describe longitudes east of the prime meridian
    # and negative numbers describe longitudes west of the prime meridian.
    longitude: Annotated[float, Ge(-180.0), Le(180.0)] | None = None
    # Defines the precision of the coordinates specified by the latitude and longitude properties.
    # This is measured in meters. The actual Location may be anywhere up to precision meters from the defined point.
    precision: float | None = None
    # The region that this Location describes.
    # The value for this property SHOULD come from the region-ov open vocabulary.
    region: str | None = None
    # The country that this Location describes.
    # This property SHOULD contain a valid ISO 3166-1 ALPHA-2 Code [ISO3166-1].
    country: str | None = None
    # The state, province, or other sub-national administrative area that this Location describes.
    # This property SHOULD contain a valid ISO 3166-2 Code [ISO3166-2].
    administrative_area: str | None = None
    # The city that this Location describes.
    city: str | None = None
    # The street address that this Location describes.
    # This property includes all aspects or parts of the street address.
    # For example, some addresses may have multiple lines including a mailstop or apartment number.
    street_address: str | None = None
    # The postal code for the Location.
    postal_code: str | None = None

    @model_validator(mode="after")
    def validate_location_properties(self) -> Self:
        """
        At least one of the following properties/sets of properties MUST be provided:
        - region
        - country
        - latitude and longitude
        """
        if (
            not self.region
            and not self.country
            and not (self.latitude and self.longitude)
        ):
            raise ValueError(
                "At least one of region, country, or both latitude and longitude MUST be provided"
            )

        if self.latitude and not self.longitude:
            raise ValueError("If latitude is present, longitude MUST be present")

        if self.longitude and not self.latitude:
            raise ValueError("If longitude is present, latitude MUST be present")

        if self.precision and (not self.latitude or not self.longitude):
            raise ValueError(
                "If precision is present, latitude and longitude MUST be present"
            )

        return self


# 4.11 Malware
class Malware(StixDomain):
    """
    Malware is a type of TTP that represents malicious code. It generally refers to a program that is inserted into a
    system, usually covertly. The intent is to compromise the confidentiality, integrity, or availability of the
    victim's data, applications, or operating system (OS) or otherwise annoy or disrupt the victim.

    The Malware SDO characterizes, identifies, and categorizes malware instances and families from data that may be
    derived from analysis. This SDO captures detailed information about how the malware works and what it does. This
    SDO captures contextual data relevant to sharing Malware data without requiring the full analysis provided by the
    Malware Analysis SDO.
    """

    type: Literal["malware"] = "malware"  # pyright: ignore[reportIncompatibleVariableOverride]
    # A name used to identify the Malware instance or family, as specified by the producer of the SDO.
    # If a name for a malware instance is not available, the SHA-256 hash value or the filename MAY be used instead.
    name: str | None = None
    # A description that provides more details and context about the Malware instance or family,
    # potentially including its purpose and its key characteristics.
    description: str | None = None
    # A set of categorizations for the malware being described.
    # The values for this property SHOULD come from the malware-type-ov open vocabulary.
    malware_types: list[str] | None = None
    # Whether the object represents a malware family (if true) or a malware instance (if false).
    is_family: bool | None = None
    # Alternative names used to identify this malware or malware family.
    aliases: list[str] | None = None
    # The list of Kill Chain Phases for which this malware can be used.
    kill_chain_phases: list[KillChainPhase] | None = None
    # The time that the malware instance or family was first seen.
    # This property is a summary property of data from sightings and other data that may or may not be available in
    # STIX. If new sightings are received that are earlier than the first seen timestamp,
    # the object may be updated to account for the new data.
    first_seen: datetime | None = None
    # The time that the malware family or malware instance was last seen.
    # This property is a summary property of data from sightings and other data that may or may not be available in
    # STIX. If new sightings are received that are later than the last_seen timestamp,
    # the object may be updated to account for the new data.
    last_seen: datetime | None = None
    # The operating systems that the malware family or malware instance is executable on.
    # This applies to virtualized operating systems as well as those running on bare metal.
    operating_system_refs: list[Identifier] | None = None
    # The processor architectures (e.g., x86, ARM, etc.) that the malware instance or family is executable on.
    # The values for this property SHOULD come from the processor-architecture-ov open vocabulary.
    architecture_execution_envs: list[str] | None = None
    # The programming language(s) used to implement the malware instance or family.
    # The values for this property SHOULD come from the implementation-language-ov open vocabulary.
    implementation_languages: list[str] | None = None
    # Any of the capabilities identified for the malware instance or family.
    # The values for this property SHOULD come from the malware-capabilities-ov open vocabulary.
    capabilities: list[str] | None = None
    # The sample_refs property specifies a list of identifiers of the SCO file or artifact objects associated with
    # this malware instance(s) or family.
    sample_refs: list[Identifier] | None = None

    @model_validator(mode="after")
    def validate_last_seen_after_first_seen(self) -> Self:
        """
        If the last_seen and the first_seen properties are both defined, then last_seen property
        MUST be greater than or equal to the timestamp in the first_seen property.
        """
        if self.first_seen and self.last_seen and self.first_seen > self.last_seen:
            raise ValueError(
                "The last_seen property MUST be greater than or equal to the timestamp in the first_seen property"
            )
        return self

    @model_validator(mode="after")
    def validate_name_if_is_family(self) -> Self:
        """
        For a malware family the name MUST be defined.
        """
        if self.is_family and not self.name:
            raise ValueError("For a malware family the name MUST be defined")
        return self


# 4.12 Malware Analysis
class MalwareAnalysis(StixDomain):
    """
    Malware Analysis captures the metadata and results of a particular static or dynamic analysis performed on a malware
    instance or family.
    """

    type: Literal["malware-analysis"] = "malware-analysis"  # pyright: ignore[reportIncompatibleVariableOverride]
    # The name of the analysis engine or product that was used.
    # Product names SHOULD be all lowercase with words separated by a dash "-".
    # For cases where the name of a product cannot be specified, a value of "anonymized" MUST be used.
    product: str
    # The version of the analysis product that was used to perform the analysis.
    version: str | None = None
    # A description of the virtual machine environment used to host the guest operating system (if applicable) that was
    # used for the dynamic analysis of the malware instance or family. If this value is not included in conjunction
    # with the operating_system_ref property, this means that the dynamic analysis may have been performed on bare
    # metal (i.e. without virtualization) or the information was redacted.
    # The value of this property MUST be the identifier for a SCO software object.
    host_vm_ref: Identifier | None = None
    # The operating system used for the dynamic analysis of the malware instance or family. This applies to
    # virtualized operating systems as well as those running on bare metal.
    # The value of this property MUST be the identifier for a SCO software object.
    operating_system_ref: Identifier | None = None
    # Any non-standard software installed on the operating system (specified through the operating-system value)
    # used for the dynamic analysis of the malware instance or family.
    # The value of this property MUST be the identifier for a SCO software object.
    installed_software_refs: list[Identifier] | None = None
    # The named configuration of additional product configuration parameters for this analysis run. For example, when
    # a product is configured to do full depth analysis of Window™ PE files. This configuration may have a named
    # version and that named version can be captured in this property. This will ensure additional runs can be
    # configured in the same way.
    configuration_version: str | None = None
    # The specific analysis modules that were used and configured in the product during this analysis run.
    # For example, configuring a product to support analysis of Dridex.
    modules: list[str] | None = None
    # The version of the analysis engine or product (including AV engines) that was used to perform the analysis.
    analysis_engine_version: str | None = None
    # The version of the analysis definitions used by the analysis tool (including AV tools).
    analysis_definition_version: str | None = None
    # The date and time that the malware was first submitted for scanning or analysis. This value will stay constant
    # while the scanned date can change. For example, when Malware was submitted to a virus analysis tool.
    submitted: datetime | None = None
    # The date and time that the malware analysis was initiated.
    analysis_started: datetime | None = None
    # The date and time that the malware analysis ended.
    analysis_ended: datetime | None = None
    # The classification result or name assigned to the malware instance by the scanner tool.
    result_name: str | None = None
    # The classification result as determined by the scanner or tool analysis process.
    # The value for this property SHOULD come from the malware-result-ov open vocabulary.
    result: str | None = None
    # This property contains the references to the STIX Cyber-observable Objects that were captured during the
    # analysis process.
    analysis_sco_refs: list[Identifier] | None = None
    # This property contains the reference to the SCO file, network traffic or artifact object that this malware
    # analysis was performed against. Caution should be observed when creating an SRO between Malware and Malware
    # Analysis objects when the Malware sample_refs property does not contain the SCO that is included in the
    # Malware Analysis sample_ref property. Note, this property can also contain a reference to an SCO which is not
    # associated with Malware (i.e., some SCO which was scanned and found to be benign.)
    sample_ref: Identifier | None = None

    @model_validator(mode="after")
    def at_least_one_of(self) -> Self:
        """
        One of result or analysis_sco_refs properties MUST be provided.
        """
        if self.result is None and self.analysis_sco_refs is None:
            raise ValueError("One of result or analysis_sco_refs must be provided")
        return self


# 4.13 Note
class Note(StixDomain):
    """
    A Note is intended to convey informative text to provide further context and/or to provide additional analysis
    not contained in the STIX Objects, Marking Definition objects, or Language Content objects which the Note relates
    to. Notes can be created by anyone (not just the original object creator).

    For example, an analyst may add a Note to a Campaign object created by another organization indicating that they've
    seen posts related to that Campaign on a hacker forum.

    Because Notes are typically (though not always) created by human analysts and are comprised of human-oriented text,
    they contain an additional property to capture the analyst(s) that created the Note. This is distinct from the
    created_by_ref property, which is meant to capture the organization that created the object.
    """

    type: Literal["note"] = "note"  # pyright: ignore[reportIncompatibleVariableOverride]
    # A brief summary of the note content.
    abstract: str | None = None
    # The content of the note.
    content: str
    # The name of the author(s) of this note (e.g., the analyst(s) that created it).
    authors: list[str] | None = None
    # The STIX Objects that the note is being applied to.
    object_refs: list[Identifier]


# 4.14 Observed Data
class ObservedData(StixDomain):
    """
    Observed Data conveys information about cyber security related entities such as files, systems, and networks using
    the STIX Cyber-observable Objects (SCOs). For example, Observed Data can capture information about an IP address,
    a network connection, a file, or a registry key. Observed Data is not an intelligence assertion, it is simply the
    raw information without any context for what it means.

    Observed Data can capture that a piece of information was seen one or more times. Meaning, it can capture both a
    single observation of a single entity (file, network connection) as well as the aggregation of multiple
    observations of an entity. When the number_observed property is 1 the Observed Data represents a single entity.
    When the number_observed property is greater than 1, the Observed Data represents several instances of an entity
    potentially collected over a period of time. If a time window is known, that can be captured using the
    first_observed and last_observed properties. When used to collect aggregate data, it is likely that some properties
    in the SCO (e.g., timestamp properties) will be omitted because they would differ for each of the individual
    observations.

    Observed Data may be used by itself (without relationships) to convey raw data collected from any source including
    analyst reports, sandboxes, and network and host-based detection tools. An intelligence producer conveying
    Observed Data SHOULD include as much context (e.g. SCOs) as possible that supports the use of the observed data set
    in systems expecting to utilize the Observed Data for improved security. This includes all SCOs that matched on an
    Indicator pattern and are represented in the collected observed event (or events) being conveyed in the
    Observed Data object. For example, a firewall could emit a single Observed Data instance containing a single
    Network Traffic object for each connection it sees. The firewall could also aggregate data and instead send out an
    Observed Data instance every ten minutes with an IP address and an appropriate number_observed value to indicate
    the number of times that IP address was observed in that window. A sandbox could emit an Observed Data instance
    containing a file hash that it discovered.

    Observed Data may also be related to other SDOs to represent raw data that is relevant to those objects.
    For example, the Sighting Relationship object, can relate an Indicator, Malware, or other SDO to a specific
    Observed Data to represent the raw information that led to the creation of the Sighting (e.g., what was actually
    seen that suggested that a particular instance of malware was active).
    """

    type: Literal["observed-data"] = "observed-data"  # pyright: ignore[reportIncompatibleVariableOverride]
    # The beginning of the time window during which the data was seen.
    first_observed: datetime
    # The end of the time window during which the data was seen.
    last_observed: datetime
    # The number of times that each Cyber-observable object represented in the objects or object_refs property was
    # seen. If present, this MUST be an integer between 1 and 999,999,999 inclusive.
    # If the number_observed property is greater than 1, the data contained in the objects or object_refs property was
    # seen multiple times. In these cases, object creators MAY omit properties of the SCO (such as timestamps) that are
    # specific to a single instance of that observed data.
    number_observed: Annotated[int, Gt(1), Gt(999999999)]
    # A dictionary of SCO representing the observation. The dictionary MUST contain at least one object.
    # The cyber observable content MAY include multiple objects if those objects are related as part of a single
    # observation. Multiple objects not related to each other via cyber observable Relationships MUST NOT be contained
    # within the same Observed Data instance.
    # For example, a Network Traffic object and two IPv4 Address objects related via the src_ref and dst_ref properties
    # can be contained in the same Observed Data because they are all related and characterize that single entity.
    # NOTE: this property is now deprecated in favor of object_refs and will be removed in a future version.
    objects: Annotated[dict[str, Any], deprecated] | None = None  # pyright: ignore[reportExplicitAny]
    # A list of SCOs and SROs representing the observation. The object_refs MUST contain at least one SCO reference
    # if defined.
    object_refs: list[Identifier] | None = None

    @model_validator(mode="after")
    def validate_last_observed_after_first_observed(self) -> Self:
        """
        The last_observed property MUST be greater than or equal to the timestamp in the first_observed property.
        """
        if (
            self.first_observed
            and self.last_observed
            and self.first_observed > self.last_observed
        ):
            raise ValueError(
                "The last_observed property MUST be greater than or equal to the the first_observed property"
            )
        return self

    @model_validator(mode="after")
    def validate_objects_or_object_refs(self) -> Self:
        """
        The objects property or the object_refs property MUST be provided,
        but both MUST NOT be present at the same time.
        """
        if self.objects is not None and self.object_refs is not None:
            raise ValueError(
                "The objects and object_refs properties MUST NOT be present at the same time"
            )
        if self.objects is None and self.object_refs is None:
            raise ValueError("Either objects or object_refs MUST be provided")
        return self


# 4.15 Opinion
class Opinion(StixDomain):
    """
    An Opinion is an assessment of the correctness of the information in a STIX Object produced by a different entity.
    The primary property is the opinion property, which captures the level of agreement or disagreement using a fixed
    scale. That fixed scale also supports a numeric mapping to allow for consistent statistical operations across
    opinions.

    For example, an analyst from a consuming organization might say that they "strongly disagree" with a Campaign
    object and provide an explanation about why. In a more automated workflow, a SOC operator might give an Indicator
    "one star" in their TIP (expressing "strongly disagree") because it is considered to be a false positive within
    their environment. Opinions are subjective, and the specification does not address how best to interpret them.
    Sharing communities are encouraged to provide clear guidelines to their constituents regarding best practice for
    the use of Opinion objects within the community.

    Because Opinions are typically (though not always) created by human analysts and are comprised of human-oriented
    text, they contain an additional property to capture the analyst(s) that created the Opinion. This is distinct
    from the created_by_ref property, which is meant to capture the organization that created the object.
    """

    type: Literal["opinion"] = "opinion"  # pyright: ignore[reportIncompatibleVariableOverride]
    # An explanation of why the producer has this Opinion. For example, if an Opinion of strongly-disagree is given,
    # the explanation can contain an explanation of why the Opinion producer disagrees and what evidence they have
    # for their disagreement.
    explanation: str | None = None
    # The name of the author(s) of this Opinion (e.g., the analyst(s) that created it).
    authors: list[str] | None = None
    # The opinion that the producer has about all of the STIX Object(s) listed in the object_refs property.
    # The values of this property MUST come from the opinion-enum enumeration.
    opinion: OpinionEnum
    # The STIX Objects that the Opinion is being applied to.
    object_refs: list[Identifier]


# 4.16 Report
class Report(StixDomain):
    """
    Reports are collections of threat intelligence focused on one or more topics, such as a description of a threat
    actor, malware, or attack technique, including context and related details. They are used to group related threat
    intelligence together so that it can be published as a comprehensive cyber threat story.

    The Report SDO contains a list of references to STIX Objects (the CTI objects included in the report) along with a
    textual description and the name of the report.

    For example, a threat report produced by ACME Defense Corp. discussing the Glass Gazelle campaign should be
    represented using Report. The Report itself would contain the narrative of the report while the Campaign SDO and
    any related SDOs (e.g., Indicators for the Campaign, Malware it uses, and the associated Relationships) would be
    referenced in the report contents.
    """

    type: Literal["report"] = "report"  # pyright: ignore[reportIncompatibleVariableOverride]
    # A name or characterizing text used to identify the Report.
    name: str
    # A description that provides more details and context about the Report, potentially including its purpose and
    # its key characteristics.
    description: str | None = None
    # The primary type(s) of content found in this report.
    # The values for this property SHOULD come from the report-type-ov open vocabulary.
    report_types: list[str] | None = None
    # The date that this Report object was officially published by the creator of this report.
    # The publication date (public release, legal release, etc.) may be different than the date the report was
    # created or shared internally (the date in the created property).
    published: datetime
    # Specifies the STIX Objects that are referred to by this Report.
    object_refs: list[Identifier]


# 4.17 Threat Actor
class ThreatActor(StixDomain):
    """
    Threat Actors are actual individuals, groups, or organizations believed to be operating with malicious intent.
    A Threat Actor is not an Intrusion Set but may support or be affiliated with various Intrusion Sets, groups, or
    organizations over time.

    Threat Actors leverage their resources, and possibly the resources of an Intrusion Set, to conduct attacks and
    run Campaigns against targets.

    Threat Actors can be characterized by their motives, capabilities, goals, sophistication level, past activities,
    resources they have access to, and their role in the organization.
    """

    type: Literal["threat-actor"] = "threat-actor"  # pyright: ignore[reportIncompatibleVariableOverride]
    # A name or characterizing text used to identify the Threat Actor or Threat Actor group.
    name: str
    # A description that provides more details and context about the Threat Actor, potentially including its purpose
    # and its key characteristics.
    description: str | None = None
    # The type(s) of this threat actor.
    # The values for this property SHOULD come from the threat-actor-type-ov open vocabulary.
    threat_actor_types: list[str] | None = None
    # A list of other names that this Threat Actor is believed to use.
    aliases: list[str] | None = None
    # The time that this Threat Actor was first seen.
    # This property is a summary property of data from sightings and other data that may or may not be available in
    # STIX. If new sightings are received that are earlier than the first seen timestamp, the object may be updated
    # to account for the new data.
    first_seen: datetime | None = None
    # The time that this Threat Actor was last seen.
    # This property is a summary property of data from sightings and other data that may or may not be available in
    # STIX. If new sightings are received that are later than the last seen timestamp, the object may be updated to
    # account for the new data
    last_seen: datetime | None = None
    # A list of roles the Threat Actor plays.
    # The values for this property SHOULD come from the threat-actor-role-ov open vocabulary.
    roles: list[str] | None = None
    # The high-level goals of this Threat Actor, namely, *what* are they trying to do.
    # For example, they may be motivated by personal gain, but their goal is to steal credit card numbers.
    # To do this, they may execute specific Campaigns that have detailed objectives like compromising point of sale
    # systems at a large retailer.
    goals: list[str] | None = None
    # The skill, specific knowledge, special training, or expertise a Threat Actor must have to perform the attack.
    # The value for this property SHOULD come from the threat-actor-sophistication-ov open vocabulary.
    sophistication: str | None = None
    # The organizational level at which this Threat Actor typically works, which in turn determines the resources
    # available to this Threat Actor for use in an attack.
    # This attribute is linked to the sophistication property — a specific resource level implies that the
    # Threat Actor has access to at least a specific sophistication level.
    # The value for this property SHOULD come from the attack-resource-level-ov open vocabulary.
    resource_level: str | None = None
    # The primary reason, motivation, or purpose behind this Threat Actor.
    # The motivation is *why* the Threat Actor wishes to achieve the goal (what they are trying to achieve).
    # The value for this property SHOULD come from the attack-motivation-ov open vocabulary.
    primary_motivation: str | None = None
    # This property specifies the secondary reasons, motivations, or purposes behind this Threat Actor.
    # These motivations can exist as an equal or near-equal cause to the primary motivation. However, it does not
    # replace or necessarily magnify the primary motivation, but it might indicate additional context.
    # The position in the list has no significance.
    # The value for this property SHOULD come from the attack-motivation-ov open vocabulary.
    secondary_motivations: list[str] | None = None
    # The personal reasons, motivations, or purposes of the Threat Actor regardless of organizational goals.
    # Personal motivation, which is independent of the organization's goals, describes what impels an individual to
    # carry out an attack. Personal motivation may align with the organization's motivation
    # — as is common with activists — but more often it supports personal goals. For example, an individual analyst may
    # join a Data Miner corporation because his or her skills may align with the corporation's objectives.
    # But the analyst most likely performs his or her daily work toward those objectives for personal reward in the
    # form of a paycheck. The motivation of personal reward may be even stronger for Threat Actors who commit illegal
    # acts, as it is more difficult for someone to cross that line purely for altruistic reasons.
    # The values for this property SHOULD come from the attack-motivation-ov open vocabulary.
    # The position in the list has no significance.
    personal_motivations: list[str] | None = None

    @model_validator(mode="after")
    def validate_last_seen_after_first_seen(self) -> Self:
        """
        If the last_seen property and the first_seen property are both defined, then the last_seen property
        MUST be greater than or equal to the timestamp in the first_seen property.
        """
        if self.first_seen and self.last_seen and self.first_seen > self.last_seen:
            raise ValueError(
                "The last_seen property MUST be greater than or equal to the timestamp in the first_seen property"
            )
        return self


# 4.18 Tool
class Tool(StixDomain):
    """
    Tools are legitimate software that can be used by threat actors to perform attacks. Knowing how and when threat
    actors use such tools can be important for understanding how campaigns are executed. Unlike malware, these tools
    or software packages are often found on a system and have legitimate purposes for power users, system
    administrators, network administrators, or even normal users. Remote access tools (e.g., RDP) and network
    scanning tools (e.g., Nmap) are examples of Tools that may be used by a Threat Actor during an attack.

    The Tool SDO characterizes the properties of these software tools and can be used as a basis for making an
    assertion about how a Threat Actor uses them during an attack. It contains properties to name and describe the
    tool, a list of Kill Chain Phases the tool can be used to carry out, and the version of the tool.

    This SDO MUST NOT be used to characterize malware. Further, Tool MUST NOT be used to characterize tools used as
    part of a course of action in response to an attack.
    """

    type: Literal["tool"] = "tool"  # pyright: ignore[reportIncompatibleVariableOverride]
    # A name or characterizing text used to identify the Tool.
    name: str
    # A description that provides more details and context about the Tool, potentially including its purpose and
    # its key characteristics.
    description: str | None = None
    # The kind(s) of tool(s) being described.
    # The values for this property SHOULD come from the tool-type-ov open vocabulary.
    tool_types: list[str] | None = None
    # Alternative names used to identify this Tool.
    aliases: list[str] | None = None
    # The list of kill chain phases for which this Tool can be used.
    kill_chain_phases: list[KillChainPhase] | None = None
    # The version identifier associated with the Tool.
    tool_version: str | None = None


# 4.19 Vulnerability
class Vulnerability(StixDomain):
    """
    A Vulnerability is a weakness or defect in the requirements, designs, or implementations of the computational
    logic (e.g., code) found in software and some hardware components (e.g., firmware) that can be directly
    exploited to negatively impact the confidentiality, integrity, or availability of that system.

    CVE is a list of information security vulnerabilities and exposures that provides common names for publicly known
    problems [CVE]. For example, if a piece of malware exploits CVE-2015-12345, a Malware object could be linked to a
    Vulnerability object that references CVE-2015-12345.

    The Vulnerability SDO is primarily used to link to external definitions of vulnerabilities or to describe 0-day
    vulnerabilities that do not yet have an external definition. Typically, other SDOs assert relationships to
    Vulnerability objects when a specific vulnerability is targeted and exploited as part of malicious cyber activity.
    As such, Vulnerability objects can be used as a linkage to the asset management and compliance process.
    """

    type: Literal["vulnerability"] = "vulnerability"  # pyright: ignore[reportIncompatibleVariableOverride]
    # A list of external references which refer to non-STIX information.
    # This property MAY be used to provide one or more Vulnerability identifiers, such as a CVE ID [CVE].
    # When specifying a CVE ID, the source_name property of the external reference MUST be set to cve and the
    # external_id property MUST be the exact CVE identifier.
    external_references: list[ExternalReference] | None = None
    # A name or characterizing text used to identify the Vulnerability.
    name: str
    # A description that provides more details and context about the Vulnerability, potentially including its
    # purpose and its key characteristics.
    description: str | None = None


SDOs = Annotated[
    (
        AttackPattern
        | Campaign
        | CourseOfAction
        | Grouping
        | Identity
        | Incident
        | Indicator
        | Infrastructure
        | IntrusionSet
        | Location
        | Malware
        | MalwareAnalysis
        | Note
        | ObservedData
        | Opinion
        | Report
        | ThreatActor
        | Tool
        | Vulnerability
    ),
    Field(discriminator="type"),
]



import ipaddress
from datetime import datetime
from annotated_types import Ge, Le
from typing_extensions import Annotated, TypedDict
from typing import Any, Literal, Self

from pydantic.functional_serializers import SerializeAsAny
from pydantic.functional_validators import model_validator
from pydantic.types import JsonValue
from pydantic import Field

from stidantic.types import (
    Extension,
    StixCore,
    StixObservable,
    StixBinary,
    StixUrl,
    Hashes,
    Identifier,
)
from stidantic.vocab import (
    EncryptionAlgorithm,
    NetworkSocketAddressFamily,
    NetworkSocketType,
    WindowsIntegrityLevel,
    WindowsServiceStartType,
    WindowsServiceStatus,
    WindowsServiceType,
    WindowsRegistryDatatype,
)


# 6.1 Artifact Object
class Artifact(StixObservable):
    """
    The Artifact Object permits capturing an array of bytes (8-bits),
    as a base64-encoded string string, or linking to a file-like payload.

    It is incumbent on sharing communities to ensure that the URL is accessible for downstream consumers.
    """

    type: Literal["artifact"] = "artifact"  # pyright: ignore[reportIncompatibleVariableOverride]
    # Whenever feasible, this value SHOULD be one of the values defined in the Template column in the IANA media type
    # registry [Media Types]. Maintaining a comprehensive universal catalog of all extant file types is obviously
    # not possible. When specifying a MIME Type not included in the IANA registry, implementers should use their best
    # judgement so as to facilitate interoperability.
    mime_type: str | None = None
    # Specifies the binary data contained in the artifact as a base64-encoded string.
    payload_bin: StixBinary | None = None
    # The value of this property MUST be a valid URL that resolves to the unencoded content.
    url: StixUrl | None = None
    # Specifies a dictionary of hashes for the contents of the url or the payload_bin.
    hashes: Hashes | None = None
    # If the artifact is encrypted, specifies the type of encryption algorithm the binary data
    # (either via payload_bin or url) is encoded in.
    # If both mime_type and encryption_algorithm are included, this signifies that the artifact represents an
    # encrypted archive.
    encryption_algorithm: EncryptionAlgorithm | None = None
    # Specifies the decryption key for the encrypted binary data (either via payload_bin or url). For example,
    # this may be useful in cases of sharing malware samples, which are often encoded in an encrypted archive.
    decryption_key: str | None = None

    @model_validator(mode="after")
    def at_least_one_of(self) -> Self:
        """
        One of payload_bin or url MUST be provided.
        """
        if self.payload_bin or self.hashes:
            return self
        raise ValueError("Missing at least hashes or payload_bin property.")

    @model_validator(mode="after")
    def url_must_not_be_present_if_payload_bin_provided(self) -> Self:
        """
        The url property MUST NOT be present if payload_bin is provided.
        """
        if self.payload_bin and self.url:
            raise ValueError(
                "The url property MUST NOT be present if payload_bin is provided"
            )
        return self

    @model_validator(mode="after")
    def hashes_must_be_present_if_url_provided(self) -> Self:
        """
        The hashes property MUST be present when the url property is present.
        """
        if self.url and not self.hashes:
            raise ValueError("The hashes property MUST be present if url is provided")
        return self

    @model_validator(mode="after")
    def decryption_key_must_not_be_present_if_encryption_algorithm_absent(self) -> Self:
        """
        The decryption_key property MUST NOT be present when the encryption_algorithm property is absent.
        """
        if not self.encryption_algorithm and self.decryption_key:
            raise ValueError(
                "The decryption_key MUST NOT be present when the encryption_algorithm property is absent"
            )
        return self

    class Config:
        json_schema_extra: dict[str, list[str]] = {
            "id_contributing_properties": ["hashes", "payload_bin"]
        }


# 6.2 Autonomous System
class AutonomousSystem(StixObservable):
    """
    The AS object represents the properties of an Autonomous Systems (AS).
    """

    type: Literal["autonomous-system"] = "autonomous-system"  # pyright: ignore[reportIncompatibleVariableOverride]
    # Specifies the number assigned to the AS. Such assignments a
    # re typically performed by a Regional Internet Registry (RIR).
    number: int
    # Specifies the name of the AS.
    name: str | None = None
    # Specifies the name of the Regional Internet Registry (RIR) that assigned the number to the AS.
    rir: str | None = None

    class Config:
        json_schema_extra: dict[str, list[str]] = {
            "id_contributing_properties": ["number"]
        }


# 6.3 Directory Object
class Directory(StixObservable):
    """
    The Directory object represents the properties common to a file system directory.
    """

    type: Literal["directory"] = "directory"  # pyright: ignore[reportIncompatibleVariableOverride]
    # Specifies the path, as originally observed, to the directory on the file system.
    path: str
    # Specifies the observed encoding for the path. The value MUST be specified if the path is stored in a
    # non-Unicode encoding. This value MUST be specified using the corresponding name from the 2013-12-20
    # revision of the IANA character set registry [Character Sets]. If the preferred MIME name for a character
    # set is defined, this value MUST be used; if it is not defined, then the Name value from the registry
    # MUST be used instead.
    path_enc: str | None = None
    # Specifies the date/time the directory was created.
    ctime: datetime | None = None
    # Specifies the date/time the directory was last written to/modified.
    mtime: datetime | None = None
    # Specifies the date/time the directory was last accessed.
    atime: datetime | None = None
    # Specifies a list of references to other File and/or Directory objects contained within the directory.
    # The objects referenced in this list MUST be of type file or directory.
    contains_refs: list[Identifier] | None = None

    class Config:
        json_schema_extra: dict[str, list[str]] = {
            "id_contributing_properties": ["path"]
        }


# 6.4 Domain Name Object
class DomainName(StixObservable):
    """
    The Domain Name object represents the properties of a network domain name.
    """

    # NOTE: As for the validation of the value field, a complex regular expression could be used to validate the
    # domain name structure. However, fully compliant regex for domain names is notoriously difficult and prone to
    # errors, especially with internationalized domain names (IDNs). A more robust solution would be to use a dedicated
    # library for domain name parsing and validation.

    type: Literal["domain-name"] = "domain-name"  # pyright: ignore[reportIncompatibleVariableOverride]
    # Specifies the value of the domain name. The value of this property MUST conform to [RFC1034], and each
    # domain and sub-domain contained within the domain name MUST conform to [RFC5890].
    value: str
    # Specifies a list of references to one or more IP addresses or domain names that the domain name resolves to.
    # The objects referenced in this list MUST be of type ipv4-addr or ipv6-addr or domain-name
    # (for cases such as CNAME records).
    resolves_to_refs: list[Identifier] | None = None

    class Config:
        json_schema_extra: dict[str, list[str]] = {
            "id_contributing_properties": ["value"]
        }


# 6.5 Email Address Object
class EmailAddress(StixObservable):
    """
    The Email Address object represents a single email address.
    """

    type: Literal["email-addr"] = "email-addr"  # pyright: ignore[reportIncompatibleVariableOverride]
    # Specifies the value of the email address. This MUST NOT include the display name.
    # This property corresponds to the addr-spec construction in section 3.4 of [RFC5322].
    value: str
    # Specifies a single email display name, i.e., the name that is displayed to the human user of a mail application.
    # This property corresponds to the display-name construction in section 3.4 of [RFC5322].
    display_name: str | None = None
    # Specifies the user account that the email address belongs to, as a reference to a User Account object.
    # The object referenced in this property MUST be of type user-account.
    belongs_to_ref: Identifier | None = None

    class Config:
        json_schema_extra: dict[str, list[str]] = {
            "id_contributing_properties": ["value"]
        }


# 6.6.2 Email MIME Component Type
class EmailMimeComponent(StixCore):
    """
    Specifies one component of a multi-part email body.

    There is no property to capture the value of the "Content-Transfer-Encoding" header field, since the body MUST be
    decoded before being represented in the body property.
    """

    # Specifies the contents of the MIME part if the content_type is not provided or starts with text/
    # (e.g., in the case of plain text or HTML email).
    # For inclusion in this property, the contents MUST be decoded to Unicode. Note that the charset provided in
    # content_type is for informational usage and not for decoding of this property.
    body: str | None = None
    # Specifies the contents of non-textual MIME parts, that is those whose content_type does not start with text/,
    # as a reference to an Artifact object or File object.
    # The object referenced in this property MUST be of type artifact or file.
    # For use cases where conveying the actual data contained in the MIME part is of primary importance,
    # artifact SHOULD be used. Otherwise, for use cases where conveying metadata about the file-like properties of the
    # MIME part is of primary importance, file SHOULD be used.
    body_raw_ref: Identifier | None = None
    # Specifies the value of the "Content-Type" header field of the MIME part.
    # Any additional "Content-Type" header field parameters such as charset SHOULD be included in this property.
    content_type: str | None = None
    # Specifies the value of the "Content-Disposition" header field of the MIME part.
    content_disposition: str | None = None

    @model_validator(mode="after")
    def validate_body_or_body_raw_ref(self) -> Self:
        """
        One of body OR body_raw_ref MUST be included.
        """
        if self.body is None and self.body_raw_ref is None:
            raise ValueError("One of body or body_raw_ref MUST be included")
        return self


# 6.6 Email Message Object
class EmailMessage(StixObservable):
    """
    The Email Message object represents an instance of an email message, corresponding to the internet message format
    described in [RFC5322] and related RFCs.

    Header field values that have been encoded as described in section 2 of [RFC2047] MUST be decoded before inclusion
    in Email Message object properties. For example, this is some text MUST be used instead of
    =?iso-8859-1?q?this=20is=20some=20text?=. Any characters in the encoded value which cannot be decoded into Unicode
    SHOULD be replaced with the 'REPLACEMENT CHARACTER' (U+FFFD). If it is necessary to capture the header value as
    observed, this can be achieved by referencing an Artifact object through the raw_email_ref property.
    """

    type: Literal["email-message"] = "email-message"  # pyright: ignore[reportIncompatibleVariableOverride]
    # Indicates whether the email body contains multiple MIME parts.
    is_multipart: bool
    # Specifies the date/time that the email message was sent.
    date: datetime | None = None
    # Specifies the value of the "Content-Type" header of the email message.
    content_type: str | None = None
    # Specifies the value of the "From:" header of the email message.
    # The "From:" field specifies the author of the message, that is, the mailbox(es) of the person or system
    # responsible for the writing of the message.
    from_ref: Identifier | None = None
    # Specifies the value of the "Sender" field of the email message.
    # The "Sender:" field specifies the mailbox of the agent responsible for the actual transmission of the message.
    sender_ref: Identifier | None = None
    # Specifies the mailboxes that are "To:" recipients of the email message.
    to_refs: list[Identifier] | None = None
    # Specifies the mailboxes that are "CC:" recipients of the email message.
    cc_refs: list[Identifier] | None = None
    # Specifies the mailboxes that are "BCC:" recipients of the email message.
    # As per [RFC5322], the absence of this property should not be interpreted as semantically equivalent to an absent
    # BCC header on the message being characterized.
    bcc_refs: list[Identifier] | None = None
    # Specifies the Message-ID field of the email message.
    message_id: str | None = None
    # Specifies the subject of the email message.
    subject: str | None = None
    # Specifies one or more "Received" header fields that may be included in the email headers.
    # List values MUST appear in the same order as present in the email message.
    received_lines: list[str] | None = None
    # Specifies any other header fields found in the email message, as a dictionary.
    # Each key/value pair in the dictionary represents the name/value of a single header field or names/values of a
    # header field that occurs more than once. Each dictionary key SHOULD be a case-preserved version of the header
    # field name. The corresponding value for each dictionary key MUST always be a list of type string to support when
    # a header field is repeated.
    additional_header_fields: dict[str, list[str]] | None = None
    # Specifies a string containing the email body.
    body: str | None = None
    # Specifies a list of the MIME parts that make up the email body.
    body_multipart: list[EmailMimeComponent] | None = None
    # Specifies the raw binary contents of the email message, including both the headers and body, as a reference
    # to an Artifact object. The object referenced in this property MUST be of type artifact.
    raw_email_ref: Identifier | None = None

    @model_validator(mode="after")
    def validate_body(self) -> Self:
        """
        The property body MUST NOT be used if is_multipart is true.
        The property body_multipart MUST NOT be used if is_multipart is false.
        """
        if self.is_multipart and self.body is not None:
            raise ValueError("body MUST NOT be used if is_multipart is true")
        if not self.is_multipart and self.body_multipart is not None:
            raise ValueError("body_multipart MUST NOT be used if is_multipart is false")
        return self

    class Config:
        json_schema_extra: dict[str, list[str]] = {
            "id_contributing_properties": ["from_ref", "subject", "body"]
        }


# 6.7.2 Archive File Extension
class ArchiveFileExtension(Extension):
    """
    The Archive File extension specifies a default extension for capturing properties specific to archive files.
    The key for this extension when used in the extensions dictionary MUST be archive-ext.
    Note that this predefined extension does not use the extension facility described in section 7.3.
    """

    # This property specifies the files that are contained in the archive. It MUST contain references to one or more
    # File objects. The objects referenced in this list MUST be of type file or directory.
    contains_refs: list[Identifier]
    # Specifies a comment included as part of the archive file.
    comment: str | None = None


# 6.7.3.2 Alternate Data Stream Type
class AlternateDataStream(StixCore):
    """
    The Alternate Data Stream type represents an NTFS alternate data stream.
    """

    # Specifies the name of the alternate data stream.
    name: str
    # Specifies a dictionary of hashes for the data contained in the alternate data stream.
    # Dictionary keys MUST come from the hash-algorithm-ov open vocabulary.
    hashes: Hashes | None = None
    # Specifies the size of the alternate data stream, in bytes. The value of this property MUST NOT be negative.
    size: Annotated[int, Ge(0)] | None = None


# 6.7.3 NTFS File Extension
class NTFSFileExtension(Extension):
    """
    The NTFS File extension specifies a set of properties specific to files stored on NTFS file systems.

    The key for this extension when used in the extensions dictionary MUST be ntfs-ext. Note that this predefined
    extension does not use the extension facility described in section 7.3.
    """

    # Specifies the security ID (SID) value assigned to the file.
    sid: str | None = None
    # Specifies a list of NTFS alternate data streams that exist for the file.
    alternate_data_streams: list[AlternateDataStream] | None = None

    @model_validator(mode="after")
    def at_least_one_of(self) -> Self:
        """
        An object using the NTFS File Extension MUST contain at least one property from this extension.
        """
        if self.sid is None and self.alternate_data_streams is None:
            raise ValueError("At least one property must be present")
        return self


# 6.7.4 PDF File Extension
class PDFFileExtension(Extension):
    """
    The PDF file extension specifies a default extension for capturing properties specific to PDF files.

    The key for this extension when used in the extensions dictionary MUST be pdf-ext.

    Note that this predefined extension does not use the extension facility described in section 7.3.
    """

    # Specifies the decimal version number of the string from the PDF header that specifies the version of the PDF
    # specification to which the PDF file conforms. E.g 1.4
    version: str | None = None
    # Specifies whether the PDF file has been optimized.
    is_optimized: bool | None = None
    # Specifies details of the PDF document information dictionary (DID), which includes properties like the document
    # creation date and producer, as a dictionary. Each key in the dictionary SHOULD be a case-preserved version of
    # the corresponding entry in the document information dictionary without the prepended forward slash, e.g., Title.
    # The corresponding value for the key MUST be the value specified for the document information dictionary entry,
    # as a string.
    document_info_dict: dict[str, str] | None = None
    # Specifies the first file identifier found for the PDF file.
    pdfid0: str | None = None
    # Specifies the second file identifier found for the PDF file.
    pdfid1: str | None = None

    @model_validator(mode="after")
    def at_least_one_of(self) -> Self:
        """
        An object using the PDF File Extension MUST contain at least one property from this extension.
        """
        if (
            self.version is None
            and self.is_optimized is None
            and self.document_info_dict is None
            and self.pdfid0 is None
            and self.pdfid1 is None
        ):
            raise ValueError("At least one property must be present")
        return self


# 6.7.5 Raster Image File Extension
class RasterImageFileExtension(Extension):
    """
    The Raster Image file extension specifies a default extension for capturing properties specific to raster image
    files.

    The key for this extension when used in the extensions dictionary MUST be raster-image-ext. Note that this
    predefined extension does not use the extension facility described in section 7.3.
    """

    # Specifies the height of the image in the image file, in pixels.
    image_height: int | None = None
    # Specifies the width of the image in the image file, in pixels.
    image_width: int | None = None
    # Specifies the sum of bits used for each color channel in the image file, and thus the total number of pixels
    # used for expressing the color depth of the image.
    bits_per_pixel: int | None = None
    # Specifies the set of EXIF tags found in the image file, as a dictionary. Each key/value pair in the dictionary
    # represents the name/value of a single EXIF tag. Accordingly, each dictionary key MUST be a case-preserved
    # version of the EXIF tag name, e.g., XResolution. Each dictionary value MUST be either an integer
    # (for int* EXIF datatypes) or a string (for all other EXIF datatypes).
    exif_tags: dict[str, int | str] | None = None

    @model_validator(mode="after")
    def at_least_one_of(self) -> Self:
        """
        An object using the Raster Image File Extension MUST contain at least one property from this extension.
        """
        if (
            self.image_height is None
            and self.image_width is None
            and self.bits_per_pixel is None
            and self.exif_tags is None
        ):
            raise ValueError("At least one property must be present")
        return self


# 6.7.6.3 Windows PE Section Type
class WindowsPESection(StixCore):
    """
    The Windows PE Section type specifies metadata about a PE file section.
    """

    # Specifies the name of the section.
    name: str
    # Specifies the size of the section, in bytes. The value of this property MUST NOT be negative.
    size: Annotated[int, Ge(0)] | None = None
    # Specifies the calculated entropy for the section, as calculated using the Shannon algorithm [Shannon Entropy].
    # The size of each input character is defined as a byte, resulting in a possible range of 0 through 8.
    entropy: float | None = None
    # Specifies any hashes computed over the section.
    hashes: Hashes | None = None


# 6.7.6.2 Windows PE Optional Header Type
class WindowsPEOptionalHeader(StixCore):
    """
    The Windows PE Optional Header type represents the properties of the PE optional header.
    An object using the Windows PE Optional Header Type MUST contain at least one property from this type.
    """

    # Specifies the hex value that indicates the type of the PE binary.
    magic_hex: str | None = None
    # Specifies the linker major version number.
    major_linker_version: int | None = None
    # Specifies the linker minor version number.
    minor_linker_version: int | None = None
    # Specifies the size of the code (text) section.
    # If there are multiple such sections, this refers to the sum of the sizes of each section.
    size_of_code: Annotated[int, Ge(0)] | None = None
    # Specifies the size of the initialized data section.
    # If there are multiple such sections, this refers to the sum of the sizes of each section.
    size_of_initialized_data: Annotated[int, Ge(0)] | None = None
    # Specifies the size of the uninitialized data section.
    # If there are multiple such sections, this refers to the sum of the sizes of each section.
    size_of_uninitialized_data: Annotated[int, Ge(0)] | None = None
    # Specifies the address of the entry point relative to the image base when the executable is loaded into memory.
    address_of_entry_point: int | None = None
    # Specifies the address that is relative to the image base of the beginning-of-code section when it is loaded
    # into memory.
    base_of_code: int | None = None
    # Specifies the address that is relative to the image base of the beginning-of-data section when it is loaded
    # into memory.
    base_of_data: int | None = None
    # Specifies the preferred address of the first byte of the image when loaded into memory.
    image_base: int | None = None
    # Specifies the alignment (in bytes) of PE sections when they are loaded into memory.
    section_alignment: int | None = None
    # Specifies the factor (in bytes) that is used to align the raw data of sections in the image file.
    file_alignment: int | None = None
    # Specifies the major version number of the required operating system.
    major_os_version: int | None = None
    # Specifies the minor version number of the required operating system.
    minor_os_version: int | None = None
    # Specifies the major version number of the image.
    major_image_version: int | None = None
    # Specifies the minor version number of the image.
    minor_image_version: int | None = None
    # Specifies the major version number of the subsystem.
    major_subsystem_version: int | None = None
    # Specifies the minor version number of the subsystem.
    minor_subsystem_version: int | None = None
    # Specifies the reserved win32 version value.
    win32_version_value_hex: str | None = None
    # Specifies the size of the image in bytes, including all headers, as the image is loaded in memory.
    size_of_image: Annotated[int, Ge(0)] | None = None
    # Specifies the combined size of the MS-DOS, PE header, and section headers, rounded up to a multiple of the
    # value specified in the file_alignment header.
    size_of_headers: Annotated[int, Ge(0)] | None = None
    # Specifies the checksum of the PE binary.
    checksum_hex: str | None = None
    # Specifies the subsystem (e.g., GUI, device driver, etc.) that is required to run this image.
    subsystem_hex: str | None = None
    # Specifies the flags that characterize the PE binary.
    dll_characteristics_hex: str | None = None
    # Specifies the size of the stack to reserve, in bytes.
    size_of_stack_reserve: Annotated[int, Ge(0)] | None = None
    # Specifies the size of the stack to commit, in bytes.
    size_of_stack_commit: Annotated[int, Ge(0)] | None = None
    # Specifies the size of the local heap space to reserve, in bytes.
    size_of_heap_reserve: Annotated[int, Ge(0)] | None = None
    # Specifies the size of the local heap space to commit, in bytes.
    size_of_heap_commit: Annotated[int, Ge(0)] | None = None
    # Specifies the reserved loader flags.
    loader_flags_hex: str | None = None
    # Specifies the number of data-directory entries in the remainder of the optional header.
    number_of_rva_and_sizes: int | None = None
    # Specifies any hashes that were computed for the optional header.
    hashes: Hashes | None = None

    @model_validator(mode="before")
    @classmethod
    def at_least_one(cls, data: Any) -> Any:  # pyright: ignore[reportExplicitAny, reportAny]
        """
        An object using the Windows PE Optional Header Type MUST contain at least one property from this type.
        """
        if isinstance(data, dict):
            for key, value in data.items():  # pyright: ignore[reportUnknownVariableType]
                if key != "type" and value is not None:
                    return data  # pyright: ignore[reportUnknownVariableType]
            raise ValueError("At least one property must be present")
        raise TypeError("Input data must be a dictionary")


# 6.7.6 Windows PE Binary File Extension
class WindowsPEBinaryExtension(Extension):
    """
    The Windows™ PE Binary File extension specifies a default extension for capturing properties specific to
    Windows portable executable (PE) files.

    The key for this extension when used in the extensions dictionary MUST be windows-pebinary-ext.
    Note that this predefined extension does not use the extension facility described in section 7.3.

    An object using the Windows™ PE Binary File Extension MUST contain at least one property other than the
    required pe_type property from this extension.
    """

    # Specifies the type of the PE binary.
    # This is an open vocabulary and values SHOULD come from the windows-pebinary-type-ov open vocabulary.
    pe_type: str
    # Specifies the special import hash, or 'imphash', calculated for the PE Binary based on its imported
    # libraries and functions.
    imphash: str | None = None
    # Specifies the type of target machine.
    machine_hex: str | None = None
    # Specifies the number of sections in the PE binary, as a non-negative integer.
    number_of_sections: Annotated[int, Ge(0)] | None = None
    # Specifies the time when the PE binary was created.
    # The timestamp value MUST be precise to the second.
    time_date_stamp: datetime | None = None
    # Specifies the file offset of the COFF symbol table.
    pointer_to_symbol_table_hex: str | None = None
    # Specifies the number of entries in the symbol table of the PE binary, as a non-negative integer.
    number_of_symbols: Annotated[int, Ge(0)] | None = None
    # Specifies the size of the optional header of the PE binary.
    size_of_optional_header: Annotated[int, Ge(0)] | None = None
    # Specifies the flags that indicate the file’s characteristics.
    characteristics_hex: str | None = None
    # Specifies any hashes that were computed for the file header.
    file_header_hashes: Hashes | None = None
    # Specifies the PE optional header of the PE binary.
    optional_header: WindowsPEOptionalHeader | None = None
    # Specifies metadata about the sections in the PE file.
    sections: list[WindowsPESection] | None = None


FileExtensions = TypedDict(
    "FileExtensions",
    {
        "archive-ext": ArchiveFileExtension,
        "ntfs-ext": NTFSFileExtension,
        "pdf-ext": PDFFileExtension,
        "raster-image-ext": RasterImageFileExtension,
        "windows-pebinary-ext": WindowsPEBinaryExtension,
    },
    total=False,
    extra_items=SerializeAsAny[Extension],
)


# 6.7 File Object
class File(StixObservable):
    """
    The File object represents the properties of a file.
    """

    type: Literal["file"] = "file"  # pyright: ignore[reportIncompatibleVariableOverride]
    # The File object defines the following extensions. In addition to these, producers MAY create their own.
    # Dictionary keys MUST use the specification defined name (examples above) or be the id of a STIX Extension object,
    # depending on the type of extension being used.
    # The corresponding dictionary values MUST contain the contents of the extension instance.
    extensions: FileExtensions | None = None  # pyright: ignore[reportIncompatibleVariableOverride]
    # Specifies a dictionary of hashes for the file.
    # When used with the Archive File Extension, this refers to the hash of the entire archive file, not its contents.
    hashes: Hashes | None = None
    # Specifies the size of the file, in bytes.
    size: Annotated[int, Ge(0)] | None = None
    # Specifies the name of the file.
    name: str | None = None
    # Specifies the observed encoding for the name of the file. This value MUST be specified using the corresponding
    # name from the 2013-12-20 revision of the IANA character set registry [Character Sets]. If the value from the
    # Preferred MIME Name column for a character set is defined, this value MUST be used; if it is not defined, then
    # the value from the Name column in the registry MUST be used instead.
    # This property allows for the capture of the original text encoding for the file name, which may be forensically
    # relevant; for example, a file on an NTFS volume whose name was created using the windows-1251 encoding, commonly
    # used for languages based on Cyrillic script.
    name_enc: str | None = None
    # Specifies the hexadecimal constant ("magic number") associated with a specific file format that corresponds
    # to the file, if applicable.
    magic_number_hex: str | None = None
    # Specifies the MIME type name specified for the file, e.g., application/msword.
    # Whenever feasible, this value SHOULD be one of the values defined in the Template column in the IANA media type
    # registry [Media Types].
    # Maintaining a comprehensive universal catalog of all extant file types is obviously not possible. When specifying
    # a MIME Type not included in the IANA registry, implementers should use their best judgement so as to facilitate
    # interoperability.
    mime_type: str | None = None
    # Specifies the date/time the file was created.
    ctime: datetime | None = None
    # Specifies the date/time the file was last written to/modified.
    mtime: datetime | None = None
    # Specifies the date/time the file was last accessed.
    atime: datetime | None = None
    # Specifies the parent directory of the file, as a reference to a Directory object.
    parent_directory_ref: Identifier | None = None
    # Specifies a list of references to other Cyber-observable Objects contained within the file, such as another file
    # that is appended to the end of the file, or an IP address that is contained somewhere in the file.
    # This is intended for use cases other than those targeted by the Archive extension.
    contains_refs: list[Identifier] | None = None
    # Specifies the content of the file, represented as an Artifact object.
    content_ref: Identifier | None = None

    @model_validator(mode="after")
    def at_least_one_of(self) -> Self:
        """
        A File object MUST contain at least one of hashes or name.
        """
        if self.hashes is None and self.name is None:
            raise ValueError("At least one of hashes or name must be present.")
        return self

    class Config:
        json_schema_extra: dict[str, list[str]] = {
            "id_contributing_properties": [
                "hashes",
                "name",
                "extensions",
                "parent_directory_ref",
            ]
        }


# 6.8 IPv4 Address Object
class IPv4Address(StixObservable):
    """
    The IPv4 Address object represents one or more IPv4 addresses expressed using CIDR notation.
    """

    type: Literal["ipv4-addr"] = "ipv4-addr"  # pyright: ignore[reportIncompatibleVariableOverride]
    # Specifies the values of one or more IPv4 addresses expressed using CIDR notation.
    # If a given IPv4 Address object represents a single IPv4 address, the CIDR /32 suffix MAY be omitted.
    value: ipaddress.IPv4Address | ipaddress.IPv4Network
    # Specifies a list of references to one or more Layer 2 Media Access Control (MAC) addresses that the IPv4
    # address resolves to.
    resolves_to_refs: list[Identifier] | None = None
    # Specifies a list of references to one or more autonomous systems (AS) that the IPv4 address belongs to.
    belongs_to_refs: list[Identifier] | None = None

    class Config:
        json_schema_extra: dict[str, list[str]] = {
            "id_contributing_properties": ["value"]
        }


# 6.9 IPv6 Address Object
class IPv6Address(StixObservable):
    """
    The IPv6 Address object represents one or more IPv6 addresses expressed using CIDR notation.
    """

    type: Literal["ipv6-addr"] = "ipv6-addr"  # pyright: ignore[reportIncompatibleVariableOverride]
    # Specifies the values of one or more IPv6 addresses expressed using CIDR notation.
    # If a given IPv6 Address object represents a single IPv6 address, the CIDR /128 suffix MAY be omitted.
    value: ipaddress.IPv6Address | ipaddress.IPv6Network
    # Specifies a list of references to one or more Layer 2 Media Access Control (MAC) addresses that the IPv6
    # address resolves to.
    resolves_to_refs: list[Identifier] | None = None
    # Specifies a list of references to one or more autonomous systems (AS) that the IPv6 address belongs to.
    belongs_to_refs: list[Identifier] | None = None

    class Config:
        json_schema_extra: dict[str, list[str]] = {
            "id_contributing_properties": ["value"]
        }


# 6.10 MAC Address Object
class MACAddress(StixObservable):
    """
    The MAC Address object represents a single Media Access Control (MAC) address.
    """

    type: Literal["mac-addr"] = "mac-addr"  # pyright: ignore[reportIncompatibleVariableOverride]
    # Specifies the value of a single MAC address.
    # The MAC address value MUST be represented as a single colon-delimited, lowercase MAC-48 address, which MUST
    # include leading zeros for each octet.
    # TODO: check the following regex: ^([0-9a-f]{2}:){5}[0-9a-f]{2}$
    value: str

    class Config:
        json_schema_extra: dict[str, list[str]] = {
            "id_contributing_properties": ["value"]
        }


# 6.11 Mutex Object
class Mutex(StixObservable):
    """
    The Mutex object represents the properties of a mutual exclusion (mutex) object.
    """

    type: Literal["mutex"] = "mutex"  # pyright: ignore[reportIncompatibleVariableOverride]
    # Specifies the name of the mutex object.
    name: str

    class Config:
        json_schema_extra: dict[str, list[str]] = {
            "id_contributing_properties": ["name"]
        }


# 6.12.2 HTTP Request Extension
class HTTPRequestExtension(Extension):
    """
    The HTTP request extension specifies a default extension for capturing network traffic properties specific to
    HTTP requests.

    The key for this extension when used in the extensions dictionary MUST be http-request-ext.
    Note that this predefined extension does not use the extension facility described in section 7.3.
    The corresponding protocol value for this extension is http.
    """

    # Specifies the HTTP method portion of the HTTP request line, as a lowercase string.
    request_method: str
    # Specifies the value (typically a resource path) portion of the HTTP request line.
    request_value: str
    # Specifies the HTTP version portion of the HTTP request line, as a lowercase string.
    request_version: str | None = None
    # Specifies all of the HTTP header fields that may be found in the HTTP client request, as a dictionary.
    request_header: dict[str, list[str]] | None = None
    # Specifies the length of the HTTP message body, if included, in bytes.
    message_body_length: int | None = None
    # Specifies the data contained in the HTTP message body, if included.
    message_body_data_ref: Identifier | None = None


# 6.12.3 ICMP Extension
class ICMPExtension(Extension):
    """
    The ICMP extension specifies a default extension for capturing network traffic properties specific to ICMP.

    The key for this extension when used in the extensions dictionary MUST be icmp-ext.
    Note that this predefined extension does not use the extension facility described in section 7.3.
    The corresponding protocol value for this extension is icmp.
    """

    # Specifies the ICMP type byte.
    icmp_type_hex: str
    # Specifies the ICMP code byte.
    icmp_code_hex: str


# 6.12.4 Network Socket Extension
class NetworkSocketExtension(Extension):
    """
    The Network Socket extension specifies a default extension for capturing network traffic properties associated
    with network sockets.

    The key for this extension when used in the extensions dictionary MUST be socket-ext.
    Note that this predefined extension does not use the extension facility described in section 7.3.
    """

    # Specifies the address family (AF_*) that the socket is configured for.
    address_family: NetworkSocketAddressFamily
    # Specifies whether the socket is in blocking mode.
    is_blocking: bool | None = None
    # Specifies whether the socket is in listening mode.
    is_listening: bool | None = None
    # Specifies any options (e.g., SO_*) that may be used by the socket, as a dictionary.
    # Each key in the dictionary SHOULD be a case-preserved version of the option name, e.g., SO_ACCEPTCONN.
    # Each key value in the dictionary MUST be the value for the corresponding options key.
    # Each dictionary value MUST be an integer. For SO_RCVTIMEO, SO_SNDTIMEO and SO_LINGER the value represents
    # the number of milliseconds. If the SO_LINGER key is present, it indicates that the SO_LINGER option is active.
    options: dict[str | int, int] | None = None
    # Specifies the type of the socket.
    socket_type: NetworkSocketType | None = None
    # Specifies the socket file descriptor value associated with the socket, as a non-negative integer.
    socket_descriptor: int | None = None
    # Specifies the handle or inode value associated with the socket.
    socket_handle: int | None = None


# 6.12.5 TCP Extension
class TCPExtension(Extension):
    """
    The TCP extension specifies a default extension for capturing network traffic properties specific to TCP.
    An object using the TCP Extension MUST contain at least one property from this extension.
    """

    # Specifies the source TCP flags, as the union of all TCP flags observed between the start of the traffic
    # (as defined by the start property) and the end of the traffic (as defined by the end property).
    # If the start and end times of the traffic are not specified, this property SHOULD be interpreted as the union
    # of all TCP flags observed over the entirety of the network traffic being reported upon.
    src_flags_hex: str | None = None
    # Specifies the destination TCP flags, as the union of all TCP flags observed between the start of the traffic
    # (as defined by the start property) and the end of the traffic (as defined by the end property).
    # If the start and end times of the traffic are not specified, this property SHOULD be interpreted as the union
    # of all TCP flags observed over the entirety of the network traffic being reported upon.
    dst_flags_hex: str | None = None

    @model_validator(mode="after")
    def at_least_one_of(self) -> Self:
        """
        An object using the TCP Extension MUST contain at least one property from this extension.
        """
        if self.src_flags_hex is None and self.dst_flags_hex is None:
            raise ValueError("At least one property must be present")
        return self


NetworkTrafficExtensions = TypedDict(
    "NetworkTrafficExtensions",
    {
        "http-request-ext": HTTPRequestExtension,
        "icmp-ext": ICMPExtension,
        "socket-ext": NetworkSocketExtension,
        "tcp-ext": TCPExtension,
    },
    total=False,
    extra_items=SerializeAsAny[Extension],
)


# 6.12 Network Traffic Object
class NetworkTraffic(StixObservable):
    """
    The Network Traffic object represents arbitrary network traffic that originates from a source and is addressed to
    a destination. The network traffic MAY or MAY NOT constitute a valid unicast, multicast, or broadcast network
    connection. This MAY also include traffic that is not established, such as a SYN flood.

    To allow for use cases where a source or destination address may be sensitive and not suitable for sharing,
    such as addresses that are internal to an organization's network, the source and destination properties
    (src_ref and dst_ref, respectively) are defined as optional in the properties table below.
    """

    type: Literal["network-traffic"] = "network-traffic"  # pyright: ignore[reportIncompatibleVariableOverride]
    # The Network Traffic object defines the following extensions. In addition to these, producers MAY create their own.
    # Dictionary keys MUST use the specification defined name (examples above) or be the id of a STIX Extension object,
    # depending on the type of extension being used.
    # The corresponding dictionary values MUST contain the contents of the extension instance.
    extensions: NetworkTrafficExtensions | None = None  # pyright: ignore[reportIncompatibleVariableOverride]
    # Specifies the date/time the network traffic was initiated, if known.
    start: datetime | None = None
    # Specifies the date/time the network traffic ended, if known.
    end: datetime | None = None
    # Indicates whether the network traffic is still ongoing.
    is_active: bool | None = None
    # Specifies the source of the network traffic, as a reference to a Cyber-observable Object.
    # The object referenced MUST be of type ipv4-addr, ipv6-addr, mac-addr, or domain-name
    # (for cases where the IP address for a domain name is unknown).
    src_ref: Identifier | None = None
    # Specifies the destination of the network traffic, as a reference to a Cyber-observable Object.
    # The object referenced MUST be of type ipv4-addr, ipv6-addr, mac-addr, or domain-name
    # (for cases where the IP address for a domain name is unknown).
    dst_ref: Identifier | None = None
    # Specifies the source port used in the network traffic, as an integer.
    src_port: Annotated[int, Ge(0), Le(65535)] | None = None
    # Specifies the destination port used in the network traffic, as an integer.
    dst_port: Annotated[int, Ge(0), Le(65535)] | None = None
    # Specifies the protocols observed in the network traffic, along with their corresponding state.
    # Protocols MUST be listed in low to high order, from outer to inner in terms of packet encapsulation.
    # That is, the protocols in the outer level of the packet, such as IP, MUST be listed first.
    # The protocol names SHOULD come from the service names defined in the Service Name column of the
    # IANA Service Name and Port Number Registry [Port Numbers].
    # In cases where there is variance in the name of a network protocol not included in the IANA Registry,
    # content producers should exercise their best judgement, and it is recommended that lowercase names be used for
    # consistency with the IANA registry.
    # If the protocol extension is present, the corresponding protocol value for that extension
    # SHOULD be listed in this property.
    protocols: list[str]
    # Specifies the number of bytes, as a positive integer, sent from the source to the destination.
    src_byte_count: Annotated[int, Ge(0)] | None = None
    # Specifies the number of bytes, as a positive integer, sent from the destination to the source.
    dst_byte_count: Annotated[int, Ge(0)] | None = None
    # Specifies the number of packets, as a positive integer, sent from the source to the destination.
    src_packets: Annotated[int, Ge(0)] | None = None
    # Specifies the number of packets, as a positive integer, sent from the destination to the source.
    dst_packets: Annotated[int, Ge(0)] | None = None
    # Specifies any IP Flow Information Export [IPFIX] data for the traffic, as a dictionary.
    # Each key/value pair in the dictionary represents the name/value of a single IPFIX element.
    # Accordingly, each dictionary key SHOULD be a case-preserved version of the IPFIX element name,
    # e.g., octetDeltaCount. Each dictionary value MUST be either an integer or a string,
    # as well as a valid IPFIX property.
    ipfix: dict[str, int | str] | None = None
    # Specifies the bytes sent from the source to the destination.
    src_payload_ref: Identifier | None = None
    # Specifies the bytes sent from the destination to the source.
    dst_payload_ref: Identifier | None = None
    # Links to other network-traffic objects encapsulated by this network-traffic object.
    encapsulates_refs: list[Identifier] | None = None
    # Links to another network-traffic object which encapsulates this object.
    encapsulated_by_ref: Identifier | None = None

    @model_validator(mode="after")
    def validate_end(self) -> Self:
        """
        If the end property and the start property are both defined, then this property
        MUST be greater than or equal to the timestamp in the start property.
        If the is_active property is true, then the end property MUST NOT be included.
        """
        if self.is_active and self.end is not None:
            raise ValueError(
                "The end property MUST NOT be included if is_active is true"
            )
        if self.start and self.end and self.start > self.end:
            raise ValueError("The end property MUST be greater than or equal to start")
        return self

    @model_validator(mode="after")
    def at_least_one_of(self) -> Self:
        """
        A Network Traffic object MUST contain the protocols property and at least one of the src_ref or dst_ref
        properties and SHOULD contain the src_port and dst_port properties.
        """
        if self.src_ref is None and self.dst_ref is None:
            raise ValueError("At least one of src_ref or dst_ref must be present")
        return self

    class Config:
        json_schema_extra: dict[str, list[str]] = {
            "id_contributing_properties": [
                "start",
                "end",
                "src_ref",
                "dst_ref",
                "src_port",
                "dst_port",
                "protocols",
                "extensions",
            ]
        }


# 6.13.2 Windows Process Extension
class WindowsProcessExtension(Extension):
    """
    The Windows Process extension specifies a default extension for capturing properties specific to Windows processes.

    The key for this extension when used in the extensions dictionary MUST be windows-process-ext.
    Note that this predefined extension does not use the extension facility described in section 7.3.
    """

    # Specifies whether Address Space Layout Randomization (ASLR) is enabled for the process.
    aslr_enabled: bool | None = None
    # Specifies whether Data Execution Prevention (DEP) is enabled for the process.
    dep_enabled: bool | None = None
    # Specifies the current priority class of the process in Windows.
    # This value SHOULD be a string that ends in _CLASS.
    priority: str | None = None
    # Specifies the Security ID (SID) value of the owner of the process.
    owner_sid: str | None = None
    # Specifies the title of the main window of the process.
    window_title: str | None = None
    # Specifies the STARTUP_INFO struct used by the process, as a dictionary.
    # Each name/value pair in the struct MUST be represented as a key/value pair in the dictionary,
    # where each key MUST be a case-preserved version of the original name.
    # For example, given a name of "lpDesktop" the corresponding key would be lpDesktop.
    startup_info: dict[str, JsonValue] | None = None
    # Specifies the Windows integrity level, or trustworthiness, of the process.
    integrity_level: WindowsIntegrityLevel | None = None

    @model_validator(mode="before")
    @classmethod
    def at_least_one(cls, data: Any) -> Any:  # pyright: ignore[reportExplicitAny, reportAny]
        """
        An object using the Windows Process Extension MUST contain at least one property from this extension.
        """
        if isinstance(data, dict):
            for key, value in data.items():  # pyright: ignore[reportUnknownVariableType]
                if key != "type" and value is not None:
                    return data  # pyright: ignore[reportUnknownVariableType]
            raise ValueError("At least one property must be present")
        raise TypeError("Input data must be a dictionary")


# 6.13.3 Windows Service Extension
class WindowsServiceExtension(Extension):
    """
    The Windows Service extension specifies a default extension for capturing properties specific to Windows services.

    The key for this extension when used in the extensions dictionary MUST be windows-service-ext.
    Note that this predefined extension does not use the extension facility described in section 7.3.
    """

    # Specifies the name of the service.
    service_name: str | None = None
    # Specifies the descriptions defined for the service.
    descriptions: list[str] | None = None
    # Specifies the display name of the service in Windows GUI controls.
    display_name: str | None = None
    # Specifies the name of the load ordering group of which the service is a member.
    group_name: str | None = None
    # Specifies the start options defined for the service.
    start_type: WindowsServiceStartType | None = None
    # Specifies the DLLs loaded by the service, as a reference to one or more File objects.
    service_dll_refs: list[Identifier] | None = None
    # Specifies the type of the service.
    service_type: WindowsServiceType | None = None
    # Specifies the current status of the service.
    service_status: WindowsServiceStatus | None = None

    @model_validator(mode="before")
    @classmethod
    def at_least_one(cls, data: Any) -> Any:  # pyright: ignore[reportExplicitAny, reportAny]
        """
        As all properties of this extension are optional, at least one of the properties defined below MUST be
        included when using this extension.
        """
        if isinstance(data, dict):
            for key, value in data.items():  # pyright: ignore[reportUnknownVariableType]
                if key != "type" and value is not None:
                    return data  # pyright: ignore[reportUnknownVariableType]
            raise ValueError("At least one property must be present")
        raise TypeError("Input data must be a dictionary")


ProcessExtensions = TypedDict(
    "ProcessExtensions",
    {
        "windows-service-ext": WindowsServiceExtension,
        "windows-process-ext": WindowsProcessExtension,
    },
    total=False,
    extra_items=SerializeAsAny[Extension],
)


# 6.13 Process Object
class Process(StixObservable):
    """
    The Process object represents common properties of an instance of a computer program as executed on an
    operating system. A Process object MUST contain at least one property (other than type) from this object
    (or one of its extensions).
    """

    type: Literal["process"] = "process"  # pyright: ignore[reportIncompatibleVariableOverride]
    # The Process object defines the following extensions. In addition to these, producers MAY create their own.
    # Dictionary keys MUST use the specification defined name (examples above) or be the id of a STIX Extension object,
    # depending on the type of extension being used.
    # The corresponding dictionary values MUST contain the contents of the extension instance.
    extensions: ProcessExtensions | None = None  # pyright: ignore[reportIncompatibleVariableOverride]
    # Specifies whether the process is hidden.
    is_hidden: bool | None = None
    # Specifies the Process ID, or PID, of the process.
    pid: int | None = None
    # Specifies the date/time at which the process was created.
    created_time: datetime | None = None
    # Specifies the current working directory of the process.
    cwd: str | None = None
    # Specifies the full command line used in executing the process, including the process name (which may be
    # specified individually via the image_ref.name property) and any arguments.
    command_line: str | None = None
    # Specifies the list of environment variables associated with the process as a dictionary.
    # Each key in the dictionary MUST be a case preserved version of the name of the environment variable,
    # and each corresponding value MUST be the environment variable value as a string.
    environment_variables: dict[str, str] | None = None
    # Specifies the list of network connections opened by the process, as a reference to one or more
    # Network Traffic objects.
    opened_connection_refs: list[Identifier] | None = None
    # Specifies the user that created the process, as a reference to a User Account object.
    creator_user_ref: Identifier | None = None
    # Specifies the executable binary that was executed as the process image, as a reference to a File object.
    image_ref: Identifier | None = None
    # Specifies the other process that spawned (i.e. is the parent of) this one, as a reference to a Process object.
    parent_ref: Identifier | None = None
    # Specifies the other processes that were spawned by (i.e. children of) this process, as a reference to one
    # or more other Process objects.
    child_refs: list[Identifier] | None = None

    @model_validator(mode="before")
    @classmethod
    def at_least_one(cls, data: Any) -> Any:  # pyright: ignore[reportExplicitAny, reportAny]
        """
        A Process object MUST contain at least one property (other than type)
        from this object (or one of its extensions).
        """
        if isinstance(data, dict):
            for key, value in data.items():  # pyright: ignore[reportUnknownVariableType]
                if key != "type" and value is not None:
                    return data  # pyright: ignore[reportUnknownVariableType]
            raise ValueError("At least one property must be present")
        raise TypeError("Input data must be a dictionary")


# 6.14 Software Object
class Software(StixObservable):
    """
    The Software object represents high-level properties associated with software, including software products.
    """

    type: Literal["software"] = "software"  # pyright: ignore[reportIncompatibleVariableOverride]
    # Specifies the name of the software.
    name: str
    # Specifies the Common Platform Enumeration (CPE) entry for the software, if available.
    # The value for this property MUST be a CPE v2.3 entry from the official NVD CPE Dictionary [NVD].
    # While the CPE dictionary does not contain entries for all software, whenever it does contain an identifier
    # for a given instance of software, this property SHOULD be present.
    cpe: str | None = None
    # Specifies the Software Identification (SWID) Tags [SWID] entry for the software, if available. The tag attribute,
    # tagId, a globally unique identifier, SHOULD be used as a proxy identifier of the tagged product.
    swid: str | None = None
    # Specifies the languages supported by the software. The value of each list member MUST be a language code
    # conformant to [RFC5646].
    languages: list[str] | None = None
    # Specifies the name of the vendor of the software.
    vendor: str | None = None
    # Specifies the version of the software.
    version: str | None = None

    class Config:
        json_schema_extra: dict[str, list[str]] = {
            "id_contributing_properties": ["name", "cpe", "swid", "vendor", "version"]
        }


# 6.15 URL Object
class URL(StixObservable):
    """
    The URL object represents the properties of a uniform resource locator (URL).
    """

    type: Literal["url"] = "url"  # pyright: ignore[reportIncompatibleVariableOverride]
    # Specifies the value of the URL. The value of this property MUST conform to [RFC3986], more specifically
    # section 1.1.3 with reference to the definition for "Uniform Resource Locator".
    value: StixUrl

    class Config:
        json_schema_extra: dict[str, list[str]] = {
            "id_contributing_properties": ["value"]
        }


# 6.16.2 UNIX Account Extension
class UnixAccountExtension(Extension):
    """
    The UNIX account extension specifies a default extension for capturing the additional information for an account
    on a UNIX system. The key for this extension when used in the extensions dictionary MUST be unix-account-ext.

    Note that this predefined extension does not use the extension facility described in Section 7.3.
    """

    # Specifies the primary group ID of the account.
    gid: int | None = None
    # Specifies a list of names of groups that the account is a member of.
    groups: list[str] | None = None
    # Specifies the home directory of the account.
    home_dir: str | None = None
    # Specifies the account’s command shell.
    shell: str | None = None

    @model_validator(mode="before")
    @classmethod
    def at_least_one(cls, data: Any) -> Any:  # pyright: ignore[reportExplicitAny, reportAny]
        """
        An object using the UNIX Account Extension MUST contain at least one property from this extension.
        """
        if isinstance(data, dict):
            for key, value in data.items():  # pyright: ignore[reportUnknownVariableType]
                if key != "type" and value is not None:
                    return data  # pyright: ignore[reportUnknownVariableType]
            raise ValueError("At least one property must be present")
        raise TypeError("Input data must be a dictionary")


UserAccountExtensions = TypedDict(
    "UserAccountExtensions",
    {"unix-account-ext": UnixAccountExtension},
    total=False,
    extra_items=SerializeAsAny[Extension],
)


# 6.16 User Account Object
class UserAccount(StixObservable):
    """
    The User Account object represents an instance of any type of user account, including but not limited to
    operating system, device, messaging service, and social media platform accounts.
    """

    type: Literal["user-account"] = "user-account"  # pyright: ignore[reportIncompatibleVariableOverride]
    # The User Account object defines the following extensions. In addition to these, producers MAY create their own.
    # Dictionary keys MUST use the specification defined name (examples above) or be the id of a STIX Extension object,
    # depending on the type of extension being used.
    # The corresponding dictionary values MUST contain the contents of the extension instance.
    extensions: UserAccountExtensions | None = None  # pyright: ignore[reportIncompatibleVariableOverride]
    # Specifies the identifier of the account. The format of the identifier depends on the system the user account is
    # maintained in, and may be a numeric ID, a GUID, an account name, an email address, etc. The user_id property
    # should be populated with whatever field is the unique identifier for the system the account is a member of.
    # For example, on UNIX systems it would be populated with the UID.
    user_id: str | None = None
    # Specifies a cleartext credential. This is only intended to be used in capturing metadata from malware analysis
    # (e.g., a hard-coded domain administrator password that the malware attempts to use for lateral movement) and
    # SHOULD NOT be used for sharing of PII.
    credential: str | None = None
    # Specifies the account login string, used in cases where the user_id property specifies something other than what
    # a user would type when they login.
    # For example, in the case of a Unix account with user_id 0, the account_login might be "root".
    account_login: str | None = None
    # Specifies the type of the account.
    # This is an open vocabulary and values SHOULD come from the account-type-ov open vocabulary.
    account_type: str | None = None
    # Specifies the display name of the account, to be shown in user interfaces, if applicable.
    # On Unix, this is equivalent to the GECOS field.
    display_name: str | None = None
    # Indicates that the account is associated with a network service or system process (daemon), not a
    # specific individual.
    is_service_account: bool | None = None
    # Specifies that the account has elevated privileges
    # (i.e., in the case of root on Unix or the Windows Administrator account).
    is_privileged: bool | None = None
    #  	Specifies that the account has the ability to escalate privileges
    # (i.e., in the case of sudo on Unix or a Windows Domain Admin account)
    can_escalate_privs: bool | None = None
    # Specifies if the account is disabled.
    is_disabled: bool | None = None
    # Specifies when the account was created.
    account_created: datetime | None = None
    # Specifies the expiration date of the account.
    account_expires: datetime | None = None
    # Specifies when the account credential was last changed.
    credential_last_changed: datetime | None = None
    # Specifies when the account was first accessed.
    account_first_login: datetime | None = None
    # Specifies when the account was last accessed.
    account_last_login: datetime | None = None

    @model_validator(mode="before")
    @classmethod
    def at_least_one(cls, data: Any) -> Any:  # pyright: ignore[reportExplicitAny, reportAny]
        """
        As all properties of this object are optional, at least one of the properties defined below
        MUST be included when using this object.
        """
        if isinstance(data, dict):
            for key, value in data.items():  # pyright: ignore[reportUnknownVariableType]
                if key != "type" and value is not None:
                    return data  # pyright: ignore[reportUnknownVariableType]
            raise ValueError("At least one property must be present")
        raise TypeError("Input data must be a dictionary")

    class Config:
        json_schema_extra: dict[str, list[str]] = {
            "id_contributing_properties": ["account_type", "user_id", "account_login"]
        }


# 6.17.2 Windows Registry Value Type
class WindowsRegistryValueType(StixCore):
    """
    The Windows Registry Value type captures the properties of a Windows Registry Key Value.
    """

    # Specifies the name of the registry value. For specifying the default value in a registry key,
    # an empty string MUST be used.
    name: str | None = None
    # Specifies the data contained in the registry value.
    data: str | None = None
    # Specifies the registry (REG_*) data type used in the registry value.
    data_type: WindowsRegistryDatatype | None = None

    @model_validator(mode="before")
    @classmethod
    def at_least_one(cls, data: Any) -> Any:  # pyright: ignore[reportExplicitAny, reportAny]
        """
        As all properties of this object are optional, at least one of the properties defined below
        MUST be included when using this object.
        """
        if isinstance(data, dict):
            for key, value in data.items():  # pyright: ignore[reportUnknownVariableType]
                if key != "type" and value is not None:
                    return data  # pyright: ignore[reportUnknownVariableType]
            raise ValueError("At least one property must be present")
        raise TypeError("Input data must be a dictionary")


# 6.17 Windows Registry Key Object
class WindowsRegistryKey(StixObservable):
    """
    The Registry Key object represents the properties of a Windows registry key.
    """

    type: Literal["windows-registry-key"] = "windows-registry-key"  # pyright: ignore[reportIncompatibleVariableOverride]
    # Specifies the full registry key including the hive.
    key: str | None = None
    # Specifies the values found under the registry key.
    # The value of the key, including the hive portion, SHOULD be case-preserved. The hive portion of the key MUST be
    # fully expanded and not truncated; e.g., HKEY_LOCAL_MACHINE must be used instead of HKLM.
    values: list[WindowsRegistryValueType] | None = None
    # Specifies the last date/time that the registry key was modified.
    modified_time: datetime | None = None
    # Specifies a reference to the user account that created the registry key.
    # The object referenced in this property MUST be of type user-account.
    creator_user_ref: Identifier | None = None
    # Specifies the number of subkeys contained under the registry key.
    number_of_subkeys: int | None = None

    @model_validator(mode="before")
    @classmethod
    def at_least_one(cls, data: Any) -> Any:  # pyright: ignore[reportExplicitAny, reportAny]
        """
        As all properties of this object are optional, at least one of the properties defined below
        MUST be included when using this object.
        """
        if isinstance(data, dict):
            for key, value in data.items():  # pyright: ignore[reportUnknownVariableType]
                if key != "type" and value is not None:
                    return data  # pyright: ignore[reportUnknownVariableType]
            raise ValueError("At least one property must be present")
        raise TypeError("Input data must be a dictionary")

    class Config:
        json_schema_extra: dict[str, list[str]] = {
            "id_contributing_properties": ["key", "values"]
        }


# 6.18.2 X.509 v3 Extensions Type
class X509v3ExtensionsType(Extension):
    """
    The X.509 v3 Extensions type captures properties associated with X.509 v3 extensions, which serve as a mechanism
    for specifying additional information such as alternative subject names.

    Note that the use of the term "extensions" in this context refers to the X.509 v3 Extensions type and is not a
    STIX Cyber Observables extension. Therefore, it is a type that describes X.509 extensions.
    """

    # Specifies a multi-valued extension which indicates whether a certificate is a CA certificate.
    # The first (mandatory) name is CA followed by TRUE or FALSE. If CA is TRUE, then an optional pathlen name
    # followed by a non-negative value can be included. Also equivalent to the object ID (OID) value of 2.5.29.19.
    basic_constraints: str | None = None
    # Specifies a namespace within which all subject names in subsequent certificates in a certification path
    # MUST be located. Also equivalent to the object ID (OID) value of 2.5.29.30.
    name_constraints: str | None = None
    # Specifies any constraints on path validation for certificates issued to CAs.
    # Also equivalent to the object ID (OID) value of 2.5.29.36.
    policy_constraints: str | None = None
    # Specifies a multi-valued extension consisting of a list of names of the permitted key usages.
    # Also equivalent to the object ID (OID) value of 2.5.29.15.
    key_usage: str | None = None
    # Specifies a list of usages indicating purposes for which the certificate public key can be used for.
    # Also equivalent to the object ID (OID) value of 2.5.29.37.
    extended_key_usage: str | None = None
    # Specifies the identifier that provides a means of identifying certificates that contain a particular public key.
    # Also equivalent to the object ID (OID) value of 2.5.29.14.
    subject_key_identifier: str | None = None
    # Specifies the identifier that provides a means of identifying the public key corresponding to the private key
    # used to sign a certificate. Also equivalent to the object ID (OID) value of 2.5.29.35.
    authority_key_identifier: str | None = None
    # Specifies the additional identities to be bound to the subject of the certificate.
    # Also equivalent to the object ID (OID) value of 2.5.29.17.
    subject_alternative_name: str | None = None
    # Specifies the additional identities to be bound to the issuer of the certificate.
    # Also equivalent to the object ID (OID) value of 2.5.29.18.
    issuer_alternative_name: str | None = None
    # Specifies the identification attributes (e.g., nationality) of the subject.
    # Also equivalent to the object ID (OID) value of 2.5.29.9.
    subject_directory_attributes: str | None = None
    # Specifies how CRL information is obtained.
    # Also equivalent to the object ID (OID) value of 2.5.29.31.
    crl_distribution_points: str | None = None
    # Specifies the number of additional certificates that may appear in the path before anyPolicy is no longer
    # permitted. Also equivalent to the object ID (OID) value of 2.5.29.54.
    inhibit_any_policy: str | None = None
    # Specifies the date on which the validity period begins for the private key, if it is different from the
    # validity period of the certificate.
    private_key_usage_period_not_before: datetime | None = None
    # Specifies the date on which the validity period ends for the private key, if it is different from the
    # validity period of the certificate.
    private_key_usage_period_not_after: datetime | None = None
    # Specifies a sequence of one or more policy information terms, each of which consists of an object identifier
    # (OID) and optional qualifiers. Also equivalent to the object ID (OID) value of 2.5.29.32.
    certificate_policies: str | None = None
    # Specifies one or more pairs of OIDs; each pair includes an issuerDomainPolicy and a subjectDomainPolicy.
    # The pairing indicates whether the issuing CA considers its issuerDomainPolicy equivalent to the subject
    # CA’s subjectDomainPolicy. Also equivalent to the object ID (OID) value of 2.5.29.33.
    policy_mappings: str | None = None

    @classmethod
    def at_least_one(cls, data: Any) -> Any:  # pyright: ignore[reportExplicitAny, reportAny]
        """
        An object using the X.509 v3 Extensions type MUST contain at least one property from this type.
        """
        if isinstance(data, dict):
            for key, value in data.items():  # pyright: ignore[reportUnknownVariableType]
                if key != "type" and value is not None:
                    return data  # pyright: ignore[reportUnknownVariableType]
            raise ValueError("At least one property must be present")
        raise TypeError("Input data must be a dictionary")


# 6.18 X.509 Certificate Object
class X509Certificate(StixObservable):
    """
    The X.509 Certificate object represents the properties of an X.509 certificate, as defined by ITU recommendation
    X.509 [X509].
    """

    type: Literal["x509-certificate"] = "x509-certificate"  # pyright: ignore[reportIncompatibleVariableOverride]
    # Specifies whether the certificate is self-signed, i.e., whether it is signed by the same entity whose
    # identity it certifies.
    is_self_signed: bool | None = None
    # Specifies any hashes that were calculated for the entire contents of the certificate.
    hashes: Hashes | None = None
    # Specifies the version of the encoded certificate.
    version: str | None = None
    # Specifies the unique identifier for the certificate, as issued by a specific Certificate Authority.
    serial_number: str | None = None
    # Specifies the name of the algorithm used to sign the certificate.
    signature_algorithm: str | None = None
    # Specifies the name of the Certificate Authority that issued the certificate.
    issuer: str | None = None
    # Specifies the date on which the certificate validity period begins.
    validity_not_before: datetime | None = None
    # Specifies the date on which the certificate validity period ends.
    validity_not_after: datetime | None = None
    # Specifies the name of the entity associated with the public key stored in the subject public key field of the
    # certificate.
    subject: str | None = None
    # Specifies the name of the algorithm with which to encrypt data being sent to the subject.
    subject_public_key_algorithm: str | None = None
    # Specifies the modulus portion of the subject's public RSA key.
    subject_public_key_modulus: str | None = None
    # Specifies the exponent portion of the subject's public RSA key, as an integer.
    subject_public_key_exponent: int | None = None
    # Specifies any standard X.509 v3 extensions that may be used in the certificate.
    x509_v3_extensions: X509v3ExtensionsType | None = None

    @classmethod
    def at_least_one(cls, data: Any) -> Any:  # pyright: ignore[reportExplicitAny, reportAny]
        """
        An X.509 Certificate object MUST contain at least one object specific property (other than type)
        from this object.
        """
        if isinstance(data, dict):
            for key, value in data.items():  # pyright: ignore[reportUnknownVariableType]
                if key != "type" and value is not None:
                    return data  # pyright: ignore[reportUnknownVariableType]
            raise ValueError("At least one property must be present")
        raise TypeError("Input data must be a dictionary")

    class Config:
        json_schema_extra: dict[str, list[str]] = {
            "id_contributing_properties": ["hashes", "serial_number"]
        }


SCOs = Annotated[
    (
        Artifact
        | AutonomousSystem
        | Directory
        | DomainName
        | EmailAddress
        | EmailMessage
        | File
        | IPv4Address
        | IPv6Address
        | MACAddress
        | Mutex
        | NetworkTraffic
        | Process
        | Software
        | URL
        | UserAccount
        | WindowsRegistryKey
        | X509Certificate
    ),
    Field(discriminator="type"),
]



from typing import Annotated
from pydantic import Field
from stidantic.types import StixCore, Identifier, StixCommon
from stidantic.sdo import SDOs
from stidantic.sco import SCOs
from stidantic.sro import SROs
from stidantic.language import LanguageContent
from stidantic.marking import MarkingDefinition
from stidantic.extension import ExtensionDefinition


# 8. Stix Bundle
class StixBundle(StixCore):
    id: Identifier
    type: str = "bundle"
    objects: list[
        Annotated[
            (
                SROs
                | SDOs
                | SCOs
                | MarkingDefinition
                | LanguageContent
                | ExtensionDefinition
            ),
            Field(discriminator="type"),
        ]
        | StixCommon
    ]
